if(!isObject(ScratchCustomBlockGroup)) new ScriptObject(ScratchCustomBlockGroup) { class = "ArrayObject"; };

/// Scratch::registerCustomBlock
///
/// Args:
///                         
/// %internalName - String : An internal name, unique to this block only.
///                         
/// %color  - RGB          : RGB Color for the block.
///                         
/// %type   - Enum         : Block type [0 = Normal, 1 = Input, 2 = Container Container]
///                         
/// %top    - Bool         : Top connector?
///                         
/// %bottom - Bool         : Bottom connector?
///                         
/// %buffer - String       : A buffer of things to display on the block. More info below.
///                        :     > Delimited by TABs.
///                        :     > Types with brackets do not require the brackets; they are optional.
///                        :     > Use "def" in place of any bracket arguments that you do not want to change. Types:
///                        :          > "string[x,y,w,h,static] <ML String>" -- Display a string. You can include ML tags in this. If static is set to 'true', the string object will not move from its initial X Y coordinates.
///                        :          > "delim[seperator_character,argtype<0=ANY,1=TEXT_ONLY,2=BLOCKS_ONLY>,x,y]" -- Add an 'argument delimiter adder' button to the block, allowing users to add more argument inputs to the block.
///                        :          > "input" -- Add an argument box. Only text can be put in this argument.
///                        :          > "blockInput" -- Add an argument box. Only input blocks can be put in this argument.
///                        :          > "anyInput" -- Add an argument box. Text and input blocks can be put in this argument. Input blocks will take priority over text.
///                         
/// %export - String       : A string representing the format of what is exported from this block. More info below.
///                        :     > Delimited by TABs.
///                        :     > Types with brackets do not require the brackets; they are optional. Types:
///                        :          > "string <String>" -- A string. This will always be the same upon export.
///                        :          > "any" -- Get the value of an argument box.
///                        :          > "any_noFormatting" -- Get the value of an argument box, and do not format them for safety.
///                        :          > "delims" -- Get the value of the rest of the delimiter argument boxes.
///                        :          > "delims_noFormatting" -- Get the value of th rest of the delimiter argument boxes, and do not format them for safety.
///                         
/// %tooltip - String      : A string that is displayed when the user hovers their mouse over the block. More info below.
///                        :     > This argument is sort of like 'printf', in the sense that it takes arguments. Arguments:
///                        :          > "%[number]" -- Get the value from argument box 'n' and display it.
///                        :          > "%@" -- Get all added delimiter blocks' values and display them.
///                        :     > Example: "Call a function %1 with the arguments %@."
///                          
/// %endToken - String     : Optional -- This is exported after '%export' is exported. Example below.
///                        :     > Example: Exports "if(%a == 1) { return; }" when endToken is "}" and block export format is "string if(\tany\tstring ) {"
///                          
/// %connectCall - String  : A function that will be called when this block is connected to another block. If this function returns false, then the block will not connect.
///			   : Arguments: %this(Your block), %other(The block we're connecting to), %side("inside" or "insidetop"(For a container container, like an IF block) or "bottom" or "top")
///			   : Example of what to put here: "echo(\"test\");" -- This will call 'echo' with the arguments '"test", %this, %other, %side'
///			   : Optional field.
///					
/// %disconnCall - String  : A function that will be called when this block is disconnected from another block.
///			   : Arguments: %this(Your block), %side("inside" or "insidetop"(For a container container, like an IF block) or "bottom" or "top")
///			   : Example of what to put here: "echo(\"test\");" -- This will call 'echo' with the arguments '"test", %this, %side'
///			   : Optional field.
///
/// @return Returns the unique ID
function Scratch::registerCustomBlock(%internalName, %color, %type, %top, %bottom, %buffer, %export, %tooltip, %endToken, %connectCall, %disconnCall) {
	%this     = ScratchCustomBlockGroup;
	%color    = mClampF(getWord(%color, 0)/255, 0, 1) SPC mClampF(getWord(%color, 1)/255, 0, 1) SPC mClampF(getWord(%color, 2)/255, 0, 1);
	%type     = mClamp(%type, 0, 2);
	%top      = mClamp(%top, 0, 1);
	%bottom   = mClamp(%top, 0, 1);
	%uniqueID = getStringCRC(%internalName);
	for(%i=0;%i<%this.size();%i++) {
		%ID = getRecord(%this.at(%i), 0);
		if(%ID $= %uniqueID) {
			error("ERROR: Scratch::registerCustomBlock() - Block is already registered. Ignoring.");
			return 0;
		}
	}
	%this.pushBack(%uniqueID NL %internalName NL %color NL %type NL %top NL %bottom NL %buffer NL %export NL %tooltip NL %endToken NL %connectCall NL %disconnCall);
	return %uniqueID;
}

function Scratch::doCustomBlockCallback(%this, %type, %args) {
	%group = ScratchCustomBlockGroup;
	if(!isObject(%this)) return 0;
	if(!%this.isCustom) return 0;
	if((%uIdx = %this.uniqueIdx) < 0) {
		%uIdx  = mAbs(%uIdx);
		if(%uIdx >= ScratchDefaultCustomBlockGroup.size()) return 0;
		%data = ScratchDefaultCustomBlockGroup.at(%uIdx);
	} else {
		if(%uIdx >= %group.size()) return 0;
		%data = %group.at(%uIdx);
	}
	switch(%type) {
		case 0: // Connect callback
			%callB = getRecord(%data, 10);
			if((%info = Struct_funcCall(%callB)) $= "") return 0;
			%funcNamespace = getSafeVariableName(getRecord(%info, 0));
			%funcName      = getSafeVariableName(getRecord(%info, 1));
			%funcArgs      = getRecord(%info, 2);
			%funcArgs      = trim(%funcArgs TAB %this);
			%funcArgs      = trim(%funcArgs TAB %args);
			%evalStr       = "%ret = " @ (%funcNamespace $= "" ? "" : %funcNamespace @ "::") @ %funcName @ "(" @ strReplace(%funcArgs, "\t", ",") @ ");";
			eval(%evalStr);
			return mClamp(%ret, 0, 1);
		case 1: // Disconnect callback
			%callB = getRecord(%data, 11);
			if((%info = Struct_funcCall(%callB)) $= "") return 0;
			%funcNamespace = getSafeVariableName(getRecord(%info, 0));
			%funcName      = getSafeVariableName(getRecord(%info, 1));
			%funcArgs      = getRecord(%info, 2);
			%funcArgs      = trim(%funcArgs TAB %this);
			%funcArgs      = trim(%funcArgs TAB %args);
			%evalStr       = (%funcNamespace $= "" ? "" : %funcNamespace @ "::") @ %funcName @ "(" @ strReplace(%funcArgs, "\t", ", ") @ ");";
			eval(%evalStr);
	}
}

// Blank callbacks, to be packaged

function ScratchBlock::onConnected(%this, %other, %side) { } // Called when a block is connected to us

function ScratchBlock::onDisconnected(%this, %other, %side) { } // Called when a block is disconnected from us

function ScratchBlock::onConnect(%this, %other, %side) { return 1; } // Called when we are connected to a block. Return a boolean -- True if the connection is allowed, false if not.

function ScratchBlock::onDisconnect(%this, %other, %side) { } // Called when we are disconnected from a block

function ScratchBlock::onAdd(%this) { } // Called when we are spawned

function ScratchBlock::onRemove(%this) { } // Called when we are removed (deleted)

function ScratchBlock::onPreExport(%this) { } // Called when the exporter is prepping blocks for export

function ScratchBlock::onExport(%this) { }  // Called when the exporter 

// Load Scratch add-ons

function Scratch::loadAddons() {
	echo("--------- Loading Scratch Add-Ons ---------");
	%ptrn = "Add-Ons/Scratch_*/scratchInfo.txt";
	for(%file=findFirstFile(%ptrn);%file !$= "";%file=findNextFile(%ptrn)) {
		%addonPath = strReplace(filePath(%file), "Add-Ons/", "");
		echo("Scratch checking Add-On: " @ %addonPath);
		if(!isFile(%csFile = filePath(%file) @ "/client.cs")) {
			echo("\c4    client.cs not found - will not execute");
			continue;
		}
		eval("function initCustomBlock() { return 0; }");
		echo("\c4Loading Add-On: " @ %addonPath);
		exec(%csFile);
		if(initCustomBlock() $= "0") echo("\c2    Add-On does not have an initCustomBlock function defined.");
		echo(" ");
	}
	echo(" ");
}