
if(!isObject(ScratchHintGroup)) new SimSet(ScratchHintGroup);

function Scratch::getBlockName(%type) {
	switch(%type) {
		case 00: return "Declare Function";
		case 01: return "Call Function";
		case 02: return "Set var";
		case 03: return "Local variable";
		case 04: return "Global variable";
		case 05: return "String concantenation";
		case 06: return "Arithmetic operation";
		case 07: return "Return";
		case 08: return "[removed]";
		case 09: return "String concantenation (SPC)";
		case 10: return "String concantenation (TAB)";
		case 11: return "String concantenation (NL)";
		case 12: return "If";
		case 13: return "If end";
		case 14: return "Else if";
		case 15: return "Else";
		case 16: return "Logical Comparison";
		case 17: return "New object (INPUTBLOCK)";
		case 18: return "New object (NORMALBLOCK)";
		case 19: return "New object end";
		case 20: return "Package";
		case 21: return "Packaged function";
		case 22: return "Packaged function end";
		case 23: return "For";
		case 24: return "[removed]";
		case 25: return "While";
		case 26: return "[removed]";
		case 27: return "Set variable [INPUT]";
		case 28: return "Call object method";
		case 29: return "Object variable";
		case 30: return "Or";
		case 31: return "And";
		case 32: return "Pre-Defined Call";
		case 33: return "Pre-Defined Script";
		case 35: return "Ternary operation";
		case 36: return "New object variable";
	}
}

function Scratch::getValueType(%val) {
	if(%val $= "") return "nostring";
	if(Struct_funcCall(%val) !$= "") return "func_call";
	if((stripChars(%val, "0123456789.") $= "" || (stripChars(%val, "0123456789.") $= "-" && getSubStr(%val, 0, 1) $= "-")) && %val !$= ".") {
		return (strIPos(%val, ".") == -1 ? "int" : "float");
	} else {
		if(getSubStr(%val, 0, 1) !$= "\"" && getSubStr(%val, strLen(%val)-1, 1) !$= "\"") {
			return "string";
		}
	}
	return "error";
}

function Scratch::intToTextRep(%int) {
	// this barely works
	%lenTbl = "   hundred thousand million billion trillion quadrillion";
	%oneTbl = " ten twenty thirty fourty fifty sixty seventy eighty ninety";
	%twoTbl = " first second third fourth fifth sixth seventh eigth ninth tenth eleven twelve thirteenth fourteenth fifteenth sixteenth seventeenth eighteenth nineteenth";
	%aftTbl = " first second third fourth fifth sixth seventh eigth ninth tenth";
	if(%int < getWordCount(%twoTbl)) return getWord(%twoTbl, %int);
	%one    = getSubStr(%int, strLen(%int)-2, 1);
	%two    = getSubStr(%int, strLen(%int)-1, 1);
	%ret    = trim(getWord(%oneTbl, %one % getWordCount(%oneTbl)) SPC getWord(%aftTbl, %two % getWordCount(%aftTbl)) SPC getWord(%lenTbl, strLen(%int)));
	return %ret;
}

function Scratch::checkBlockSyntax(%this, %other, %slot) {
	if(!isObject(%this)) return;
	%count = 0;
	if(%slot > %this.realBlockArgCount && !%this.isDelim) {
		error("ERROR: Scratch::checkBlockSyntax() - Slot number out of range");
		return 0;
	} else if(%this.blockArgObject[%slot].isDelim) {
		%slot = mClamp(%slot, 0, %this.realBlockArgCount-1);
		echo(%this.blockArgObject[%slot].validChecks);
	}
	if(isObject(%other)) %value = Scratch::getValue(%other);
	else %value = %other;
	%checkStr   = %value;
	%valueType  = Scratch::getValueType(%checkStr);
	%validCheck = %this.blockArgObject[%slot].validChecks;
	%valid      = 1;
	for(%i=0;%i<getWordCount(%validCheck);%i++) {
		%action  = getWord(%validCheck, %i);
		%split   = Scratch::splitStatus(%action);
		%actName = getField(%split, 0);
		%actArgs = getField(%split, 1);
		%actArgs = strReplace(strReplace(strReplace(strReplace(%actArgs, "~", "\t"), "\\_", "[UNDERSCORE]"), "_", " "), "[UNDERSCORE]", "_");
		switch$(%actName) {
			case "any":    %valid = 1;
			case "string": if(%valueType !$= "string") %valid = "Expected a string";
			case "number": if(%valueType !$= "int" && %valueType !$= "float") %valid = "Expect a number";
			case "int":    if(%valueType !$= "int") %valid = "Expected an integer";
			case "float":  if(%valueType !$= "float") %valid = "Expected a float";
			case "block": 
				if(!isObject(%other) || strIPos(%actArgs, " " @ %other.blockLayoutType @ " ") == -1) {
					%bufStr = "";
					for(%j=0;%j<getWordCount(%actArgs);%j++) {
						if(getWord(%actArgs, %j) $= "") continue;
						%bufStr = (%bufStr $= "" ? "" : %bufStr @ (%j == getWordCount(%actArgs)-1 ? ", or " : ", ")) @ Scratch::getBlockName(getWord(%actArgs, %j));
					}
					%valid = "Expected a" @ (strIPos("aeiou", getSubStr(%bufStr, 0, 1)) == -1 ? "" : "n") SPC %bufStr SPC "";
				}
			case "setCheckStr":
				%start      = getField(%actArgs, 0);
				%strLen     = getField(%actArgs, 1);
				if(%strLen $= "~") %strLen = strLen(%value);
				%checkStr   = getSubStr(%value, %start, %strLen);
				%valueType  = Scratch::getValueType(%value);
			case "only_at":
				%start      = getField(%actArgs, 0);
				%strLen     = getField(%actArgs, 1);
				%stripChars = getField(%actArgs, 2);
				%sourceStr  = getSubStr(%checkStr, %start, %strLen);
				if(stripChars(%sourceStr, %stripChars) !$= "") %valid = "The " @ Scratch::intToTextRep(%start+1) @ " character can only contain any of these characters: \"" @ %stripChars @ "\"";
			case "no_at":
				%start      = getField(%actArgs, 0);
				%strLen     = getField(%actArgs, 1);
				%stripChars = getField(%actArgs, 2);
				%sourceStr  = getSubStr(%checkStr, %start, %strLen);
				if(stripChars(%sourceStr, %stripChars) !$= %sourceStr) %valid = "The " @ Scratch::intToTextRep(%start+1) @ " character cannot contain any of these characters: \"" @ %stripChars @ "\"";
			case "no":
				%stripChars = getField(%actArgs, 0);
				if(stripChars(%checkStr, %stripChars) !$= %checkStr) %valid = "Argument # " @ %slot @ " of this block cannot contain any of these characters: \"" @ %stripChars @ "\"";
			case "only":
				%stripChars = getField(%actArgs, 0);
				if(stripChars(%checkStr, %stripChars) !$= "") %valid = "Argument # " @ %slot @ " of this block can only contain these characters: \"" @ %stripChars @ "\"";
			case "notempty": if(%checkStr $= "") %valid = "Argument # " @ %slot @ " of this block cannot be empty.";
			case "safevarname": if(getSafeVariableName(%checkStr) !$= %checkStr) %valid = "Argument # " @ %slot @ " contains an unsafe variable name. Variables can only contain numbers, letters, and underscores. Variable names cannot start with a number.";
			case "canbeempty": if(%checkStr $= "") %valid = 1;
		}
	}
	return %valid;
}

function Scratch::checkScriptForErrors() {
	%group = ScratchContainer;
	for(%i=0;%i<%group.getCount();%i++) {
		%object = %group.getObject(%i);
		%error  = "";
		if(%object.blockType $= "INPUTBLOCK" && !isObject(%object.innerConnection_Parent)) continue;
		if(%error !$= "") {
			%centerPt = %object.getLeft()+(%object.getWidth()/2) SPC %object.getTop()+(%object.getHeight()/2);
			%centerPt = vectorAdd(%centerPt, Scratch_Content.getGroup().getPosition());
			%centerPt = getWords(vectorAdd(%centerPt, Scratch_MainWindow.getPosition()), 0, 1);
			Scratch::blinkError(%object);
			Scratch::debugger_error("ERROR:" SPC %error);
			Scratch::pushTooltip(%error, 1);
			Scratch::renderTooltip(1);
			Scratch_ToolTip.resize(getWord(%centerPt, 0), getWord(%centerPt, 1), Scratch_ToolTip.getWidth(), Scratch_ToolTip.getHeight());
			return 0;
		}
		for(%j=0;%j<%object.blockArgCount;%j++) {
			%blockArg = %object.blockArgObject[%j];
			%centerPt = %blockArg.getLeft()+%blockArg.getWidth()/2 SPC %blockArg.getTop()+%blockArg.getHeight()/2;
			%centerPt = vectorAdd(%centerPt, %object.getPosition());
			%centerPt = vectorAdd(%centerPt, Scratch_Content.getGroup().getPosition());
			%centerPt = getWords(vectorAdd(%centerPt, Scratch_MainWindow.getPosition()), 0, 1);
			if(isObject(%child = %blockArg.innerConnection_Child)) {
				%error = Scratch::checkBlockSyntax(%object, %child, %j);
			} else if(isObject(%child = %blockArg.textEdit)) {
				%error = Scratch::checkBlockSyntax(%object, %child.getValue(), %j);
			} else {
				%error = Scratch::checkBlockSyntax(%object, "", %j);
			}
			if(%error !$= "1") {
				Scratch::blinkError(%object);
				Scratch::debugger_error("ERROR:" SPC %error);
				Scratch::pushTooltip(%error, 1);
				Scratch::renderTooltip(1);
				Scratch_ToolTip.resize(getWord(%centerPt, 0), getWord(%centerPt, 1), Scratch_ToolTip.getWidth(), Scratch_ToolTip.getHeight());
				return 0;
			}
		}
	}
	return 1;
}