function Scratch::doesContainerHave(%this, %object) {
	if(!isObject(%this)) return 0;
	if(!isObject(%object)) return 0;
	if(!%this.container) return 0;
	%obj = %this.inside_topConnection;
	while(isObject(%obj)) {
		if(%obj == %object) return 1;
		if(%obj.container) if(Scratch::doesContainerHave(%obj, %object)) return 1;
		%obj = %obj.bottomConnection;
	}
	return 0;
}

function Scratch::attachBlock(%this, %other, %side) {
	if(%this.container) if(Scratch::doesContainerHave(%this, %other)) return 0;
	if(%other.container) if(Scratch::doesContainerHave(%other, %this)) return 0;
	Scratch::debugEcho("=================== CONNECTING " @ %this @ " TO " @ %other @ " ON SIDE \"" @ %side @ "\"");
	if(!isObject(%this)) return 0;
	if(!isObject(%other)) return 0;
	// Do a hard-coded check for some blocks
	switch(%this.blockLayoutType) {
		case 36: // NEW OBJECT VARIABLE BLOCK
			if(%other.blockLayoutType != 18) {
				%absPos   = Scratch::tut_getAbsolutePosition(%this);
				%absPos   = vectorAdd(%absPos, %this.getWidth()/2 SPC %this.getHeight()/2);
				Scratch::pushTooltip("");
				Scratch::pushTooltip("This block can only be placed inside of a NEW OBJECT block.", 0, getWord(%absPos, 0), getWord(%absPos, 1));
				return 0;
			}
	}
	switch(%other.blockLayoutType) {
		case 18: // NEW OBJECT VARIABLE BLOCK
			if(%this.blockLayoutType != 36) {
				%absPos   = Scratch::tut_getAbsolutePosition(%other);
				%absPos   = vectorAdd(%absPos, %other.getWidth()/2 SPC %other.getHeight()/2);
				Scratch::pushTooltip("");
				Scratch::pushTooltip("You can only place new object variable blocks in this block.", 0, getWord(%absPos, 0), getWord(%absPos, 1));
				return 0;
			}
	}
	// Now on to the actual script
	switch$(%side) {
		case "top": // Attach %this to the top of %other
			if(!%other.top) return 0;
			if(!%this.bottom) return 0;
			if(isObject(%other.topConnection)) return 0;
			if(%this.isCustom && !Scratch::doCustomBlockCallback(%this, 0, %other TAB %side)) return 0;
			%this.resize(%other.getLeft(), %other.getTop()-getWord(%this.getExtent(), 1), getWord(%this.getExtent(), 0), getWord(%this.getExtent(), 1));
			%other.topConnection   = %this;
			%this.bottomConnection = %other;
			%this.status = "FOLLOW[" @ Scratch::getHighestTopConnection(%other) @ "]";
			Scratch::update(Scratch::getHighestTopConnection(%other), 2);
		case "bottom":// Attach %this to the bottom of %other
			if(!%other.bottom) return 0;
			if(!%this.top) return 0;
			if(isObject(%other.bottomConnection)) return 0;
			if(isObject(%this.topConnection)) return 0;
			if(%this.isCustom && !Scratch::doCustomBlockCallback(%this, 0, %other TAB %side)) return 0;
			%this.resize(%other.getLeft(), %other.getBottom()-7, getWord(%this.getExtent(), 0), getWord(%this.getExtent(), 1));
			%this.topConnection = %other;
			%other.bottomConnection = %this;
			%this.status = "FOLLOW[" @ Scratch::getHighestTopConnection(%other) @ "]";
			Scratch::update(Scratch::getHighestTopConnection(%other));
		case "inside":
			if(%this.blockType !$= "INPUTBLOCK") return 0;
			if(%other.argType !$= "BLOCKINPUT" && %other.argType !$= "ANYINPUT") return 0;
			if(isObject(%other.innerConnection_Child)) return 0;
			if(isObject(%this.innerConnection_Parent)) return 0;
			if(%this.isCustom && !Scratch::doCustomBlockCallback(%this, 0, %other TAB %side)) return 0;
			%other.innerConnection_Child = %this;
			%this.innerConnection_Parent = %other;
			%other.getGroup().innerConnected[%this] = 1;
			%other.getGroup().innerConnectionTable = trim(%other.getGroup().innerConnectionTable SPC %this);
			%this.status = "FOLLOW[" @ %other @ "]";
			Scratch::renderBlock(Scratch::getHighestInnerConnection(%other.getGroup()));
			Scratch::renderBlock(Scratch::getHighestInnerConnection(%other.getGroup()));
			%other = %other.getGroup();
		case "insidetop":
			if(isObject(%this.topConnection)) return 0;
			if(isObject(%other.inside_topConnection)) return 0;
			if(!%this.top) return 0;
			if(%this.isCustom && !Scratch::doCustomBlockCallback(%this, 0, %other TAB %side)) return 0;
			%this.topConnection = %other;
			%other.inside_topConnection = %this;
			%this.status = "FOLLOW[" @ Scratch::getHighestTopConnection(%other) @ "]";
			Scratch::renderBlock(Scratch::getHighestTopConnection(%other));
			Scratch::renderBlock(Scratch::getHighestTopConnection(%this));
		default:
			error("ERROR: Scratch::attachBlock() - Invalid side '" @ %side @ "'."); return 0;
	}
	alxPlay(ScratchSounds_ClipIn);
	return 1;
}

function Scratch::detachBlock(%this, %side) {
	Scratch::debugEcho("=================== DISCONNECTING " @ %this @ " FROM SIDE \"" @ %side @ "\"");
	//echo("DETACH:" @ %this SPC %other);
	if(!isObject(%this)) return 0;
	switch$(%side) {
		case "top":
			if(!isObject(%other = %this.topConnection)) return 0;
			if(%other.bottomConnection != %this && %other.inside_topConnection == %this) return 0;
			%other.bottomConnection = "";
			%this.topConnection     = "";
			%this.status = "UPDATE";
			// Callbacks
			Scratch::doCustomBlockCallback(%this, 1, %other TAB %side);
		case "bottom":
			if(!isObject(%other = %this.bottomConnection)) return 0;
			%other.topConnection   = "";
			%this.bottomConnection = "";
			%this.status = "UPDATE";
			// Callbacks
			Scratch::doCustomBlockCallback(%this, 1, %other TAB %side);
		case "inside":
			if(!isObject(%other = %this.innerConnection_Parent)) return 0;
			%other.innerConnection_Child = "";
			%this.innerConnection_Parent  = "";
			%this.getGroup().lastInnerConnection = %other;
			%other.resize(%other.getLeft(), %other.getTop(), 29, 24);
			%other.getGroup().innerConnected[%this] = 0;
			%this.status = "UPDATE";
			%other.getGroup().innerConnectionTable = removeWord(%other.getGroup().innerConnectionTable, wordPos(%other.getGroup().innerConnectionTable, %this));
			Scratch::renderBlock(Scratch::getHighestInnerConnection(%other.getGroup()));
			Scratch::renderBlock(Scratch::getHighestInnerConnection(%other.getGroup()));
			Scratch::renderBlock(Scratch::getHighestInnerConnection(%this));
			Scratch::renderBlock(Scratch::getHighestInnerConnection(%this));
			Scratch_Content.pushToBack(%this);
			// Callbacks
			Scratch::doCustomBlockCallback(%this, 1, %other.getGroup() TAB %side);
		case "insidetop":
			if(!isObject(%other = %this.topConnection)) return 0;
			if(!isObject(%other.inside_topConnection)) return 0;
			%this.topConnection = "";
			%other.inside_topConnection = "";
			%this.status = "UPDATE";
			Scratch::renderBlock(Scratch::getHighestTopConnection(%other));
			Scratch::renderBlock(Scratch::getHighestTopConnection(%this));
			// Callbacks
			Scratch::doCustomBlockCallback(%this, 1, %other TAB %side);
		default:
			error("ERROR: Scratch::detachBlock() - Invalid side '" @ %side @ "'."); return 0;
	}
	alxPlay(ScratchSounds_ClipOut);
	return 1;
}

function Scratch::getHighestTopConnection(%this) {
	if(!isObject(%this)) return 0;
	%object = %this.topConnection;
	if(!isObject(%object)) return %this;
	while(isObject(%object)) {
		if(%dbg++ >= 100) { warn("WARNING: Scratch::getHighestTopConnection() - Loop limit exceeded [connections.cs]"); break; }
		if(!isObject(%object.topConnection)) return %object;
		if(%got[%object]) return %object;
		%got[%object] = 1;
		%object = %object.topConnection;
	}
}

function Scratch::getHighestInnerConnection(%this, %getChild) {
	if(!isObject(%this)) return 0;
	%object = %this.innerConnection_Parent;
	if(!isObject(%object)) return %this;
	while(isObject(%object)) {
		if(!isObject(%object.getGroup().innerConnection_Parent)) return (%getChild ? %object : %object.getGroup());
		%object = %object.getGroup().innerConnection_Parent;
	}
}

function Scratch::getInnerConnectionExtent(%this) {
	// %this has to be a block.
	if(!isObject(%this = %this.getID())) return %this.getExtent();
	%groups    = %this;
	%extent    = %groups.getExtent();
	while(%groups !$= "") {
		if(%dbg++ >= 1000) { echo("Scratch::getInnerConnectionExtent() - Borked"); break; }
		for(%i=0;%i<getWord(%groups, 0).getCount();%i++) {
			%obj = getWord(%groups, 0).getObject(%i);
			if(%obj.ignore) continue;
			if(%obj.argType !$= "ANYINPUT" && %obj.argType !$= "BLOCKINPUT") continue;
			if(!isObject(%obj.innerConnection_Child)) continue;
			%groups = %groups SPC %obj.innerConnection_Child;
			%extent = getWords(vectorAdd(%extent, %obj.innerConnection_Child.getExtent()), 0, 1);
		}
		%groups = removeWord(%groups, 0);
	}
	return %extent;
}

function Scratch::isInnerConnected(%this, %object) {
	// %object has to be a block.
	if(!isObject(%this = %this.getID())) return;
	if(!isObject(%object = %object.getID())) return;
	%groups = %this;
	while(%groups !$= "") {
		if(%dbg++ >= 1000) { echo("Borked"); break; }
		if(getWord(%groups, 0).innerConnected[%object]) return 1;
		//if(getWord(%groups, 0).getID() == %object.getID()) return 2;
		for(%i=0;%i<getWord(%groups, 0).getCount();%i++) {
			%obj = getWord(%groups, 0).getObject(%i);
			if(%obj.ignore) continue;
			if(%obj.argType !$= "ANYINPUT" && %obj.argType !$= "BLOCKINPUT") continue;
			if(!isObject(%obj.innerConnection_Child)) continue;
			%groups = %groups SPC %obj.innerConnection_Child;
		}
		%groups = removeWord(%groups, 0);
	}
	%obj = %object;
	while(isObject(%obj)) {
		if(%dbg++ >= 1000) { echo("Borked"); break; }
		if(%obj.innerConnected[%this]) return 1;
		%obj = %obj.innerConnection_Parent;
	}
	%dbg = 0;
	return 0;
}

function ScratchBlockMouseEvent::findNearestConnectionPoint(%this) {
	%this         = %this.getGroup();
	%centerPoint  = mFloor(%this.getLeft()+(getWord(%this.getExtent(), 0)/2)) SPC mFloor(%this.getTop()+(getWord(%this.getExtent(), 1)/2));
	%curX         = getWord(Canvas.getCursorPos(), 0);
	%curY         = getWord(Canvas.getCursorPos(), 1);
	%offX         = Scratch_Content.getGroup().getLeft();
	%offX        += Scratch_Content.getGroup().getGroup().getLeft();
	%offY         = Scratch_Content.getGroup().getTop();
	%offY        += Scratch_Content.getGroup().getGroup().getTop();
	%relativeCurX = mAbs(%curX-%offX);
	%relativeCurY = mAbs(%curY-%offY);
	%left         = %this.getLeft();
	%right        = %this.getRight();
	%top          = %this.getTop();
	%bottom       = %this.getBottom();
	%extraSpaceY  = 20;
	%extraSpaceX  = 40;
	%closestScore = 9195921;
	for(%i=0;%i<ScratchContainer.getCount();%i++) {
		%obj            = ScratchContainer.getObject(%i);
		%objCenterPoint = mFloor(%obj.getLeft()+(getWord(%obj.getExtent(), 0)/2)) SPC mFloor(%obj.getTop()+(getWord(%obj.getExtent(), 1)/2));
		if(%obj == %this) continue;
		if(%this.blockType $= "INPUTBLOCK") {
			if(Scratch::isInnerConnected(%this, %obj)) continue;
			for(%j=0;%j<%obj.getCount();%j++) {
				%child = %obj.getObject(%j);
				if(%child.ignore) continue;
				if(%child.argType !$= "BLOCKINPUT" && %child.argType !$= "ANYINPUT") continue;
				if(isObject(%child.innerConnection_Child)) continue;
				if(%this.getRight() >= %obj.getLeft()+%child.getLeft() && %this.getLeft() <= %obj.getLeft()+%child.getRight() && %this.getTop() <= %obj.getTop()+%child.getBottom() && %this.getBottom() >= %obj.getTop()+%child.getTop()) {
					%center = %obj.getLeft()+(%child.getLeft()+(%child.getWidth()/2)) SPC %obj.getTop()+(%child.getTop()+(%child.getHeight()/2));
					if((%score = vectorDist(%center, %relativeCurX SPC %relativeCurY)) >= %closestScore) continue;
					%closestScore = %score;
					%toRet        = "inside" TAB %child TAB %obj.getLeft()+%child.getLeft() SPC %obj.getTop()+%child.getTop();
					Scratch::renderDebugRect("Scratch_ConnectionRect", %obj.getLeft()+%child.getLeft(), %obj.getLeft()+%child.getRight(), %obj.getTop()+%child.getBottom(), %obj.getTop()+%child.getTop());
					continue;
				}
			}
		}
		if(%right >= %obj.getLeft() && %left <= %obj.getLeft()+%extraSpaceX && %top <= %obj.getBottom()+%extraSpaceY && %bottom >= %obj.getTop()-%extraSpaceY) {
			if(%obj.container) {
				%barR  = (%obj.barConnector.getRight()+%obj.getLeft())+%offX;
				%cBtm  = (%obj.container0.getBottom()+%obj.getTop()-7)+%offY;
				%cTop  = (%obj.container1.getTop()+%obj.getTop()+7)+%offX;
				if(%curX >= %barR && %curX <= (%offX+%obj.getRight()) && %curY >= %cBtm && %top <= %cTop) {
					%center = %obj.getLeft()+(%obj.getWidth()/2) SPC %cTop - ((%cBtm-%cTop)/2);
					if((%score = vectorDist(%center, %relativeCurX SPC %relativeCurY)) >= %closestScore) continue;
					%closestScore = %score;
					%toRet = "inside_top" TAB %obj TAB %barR TAB %cBtm;
					Scratch::renderDebugRect("Scratch_ConnectionRect", %barR, %cTop, %obj.getWidth(), %cBtm-%cTop);
					continue;
				}
			}
			Scratch::renderDebugRect("Scratch_ConnectionRect", %obj.getLeft(), %obj.getTop()-%extraSpaceY, %obj.getWidth(), %obj.getHeight()+(%extraSpaceY*2));
			if(getWord(%centerPoint, 1) > getWord(%objCenterPoint, 1)) {
				// %this is below the object
				if(!%obj.bottom) continue;
				if(!%this.top) continue;
				if(isObject(%obj.bottomConnection)) continue;
				%center = %obj.getLeft()+(%obj.getWidth()/2) SPC %obj.getBottom()+3;
				if((%score = vectorDist(%center, %relativeCurX SPC %relativeCurY)) >= %closestScore) continue;
				%closestScore = %score;
				%toRet = "front" TAB %obj TAB %obj.getLeft() SPC %obj.getBottom()-7;
				continue;
			} else {
				// %this is above the object
				if(isObject(%obj.topConnection)) continue;
				if(isObject(%this.bottomConnection)) continue;
				if(!%obj.top) continue;
				if(!%this.bottom) continue;
				%center = %obj.getLeft()+(%obj.getWidth()/2) SPC %obj.getTop()-5;
				if((%score = vectorDist(%center, %relativeCurX SPC %relativeCurY)) >= %closestScore) continue;
				%closestScore = %score;
				%toRet = "back" TAB %obj TAB %obj.getLeft() SPC %obj.getTop()-getWord(%this.getExtent(), 1)+7;
				continue;
			}
		}
	}
	if($DEBUG::SCRATCH::LASTRET !$= getFields(%toRet, 0, 1)) {
		$DEBUG::SCRATCH::LASTRET = getFields(%toRet, 0, 1);
		switch$(getField(%toRet, 0)) {
			case "front": %side = "bottom";
			case "back": %side = "top";
			case "inside_top": %side = "inside_top";
			case "inside": %side = "inside";
		}
		if(getField(%toRet, 1) !$= "") Scratch::debugEcho(" Got possible connection: " @ getField(%toRet, 1) @ " on side \"" @ strUpr(%side) @ "\"");
	}
	return %toRet;
}

function ScratchBlockMouseEvent::onRightMouseDown(%this) {
	if(%this.getGroup().hybridInfo !$= "") %extraBits = "Convert Block";
	Scratch::openRightClickMenu("Scratch_performBlockAction(\"" @ %this.getGroup().getID() @ "\");", "Delete\tDuplicate" @ (%extraBits !$= "" ? "\t" @ %extraBits : ""));
}

function ScratchBlockMouseEvent::onMouseDown(%this) {
	Scratch::debugEcho("+ Dragging " @ %this);
	$CACHE::Scratch::BlockSystem::LastDraggedMouseEvent = %this;
	%mousePoint                  = getWord(Canvas.getCursorPos(), 0)-getWord(%this.getGroup().getPosition(), 0) SPC getWord(Canvas.getCursorPos(), 1)-getWord(%this.getGroup().getPosition(), 1);
	%this.getGroup().mouseOffset = getWords(vectorSub(%mousePoint, vectorSub(Scratch_Content.getPosition(), "1 1 0")), 0, 1);
	%this.getGroup().getGroup().pushToBack(%this.getGroup());
	if(isObject(%obj = %this.getGroup().topConnection)) {
		if(isObject(%obj.inside_topConnection) && %obj.inside_topConnection.getId() == %this.getGroup().getId()) Scratch::detachBlock(%this.getGroup(), "insidetop");
		else Scratch::detachBlock(%this.getGroup(), "top");
	}
	if(isObject(%obj = %this.getGroup().innerConnection_Parent)) {
		%innerMost = Scratch::getHighestInnerConnection(%this.getGroup());
		Scratch::detachBlock(%this.getGroup(), "inside");
		%obj.getGroup().resize(%obj.getGroup().getLeft(), %obj.getGroup().getTop(), 30, getWord(%obj.getGroup().getExtent(), 1)); // Make sure that the block is sized correctly
	}
	%this.getGroup().status = "DRAG";
	Scratch::update(%this.getGroup());
}

function ScratchBlockMouseEvent::onMouseUp(%this) {
	Scratch::debugEcho("-");
	if(ScratchHighlightContainer.getCount() != 0) ScratchHighlightContainer.deleteAll();
	if(%this.lastConnectPoint !$= "") {
		%lyr = getField(%this.lastConnectPoint, 0);
		%obj = getField(%this.lastConnectPoint, 1);
		%pos = getField(%this.lastConnectPoint, 2);
		%x   = getWord(%pos, 0);
		%y   = getWord(%pos, 1);
		%this.getGroup().resize(%x, %y, getWord(%this.getGroup().getExtent(), 0), getWord(%this.getGroup().getExtent(), 1));
		%gui = %this.getGroup().getGroup();
		if(%lyr $= "back") {
			// Above
			Scratch::attachBlock(%this.getGroup(), %obj, "top");
			%this.getGroup().status = "READY";
		} else if(%lyr $= "front") {
			// Below
			Scratch::attachBlock(%this.getGroup(), %obj, "bottom");
			%this.getGroup().status = "READY";
		} else if(%lyr $= "inside") {
			// Inside
			Scratch::attachBlock(%this.getGroup(), %obj, "inside");
		} else if(%lyr $= "inside_top") {
			Scratch::attachBlock(%this.getGroup(), %obj, "insidetop");
		} else %this.getGroup().status = "READY";
	} else %this.getGroup().status = "READY";
	Scratch::update(%this.getGroup());
	//cancel(%this.dragLoop);
}