if(!isObject(ScratchFuncHelp)) new ScriptObject(ScratchFuncHelp) { class = "ArrayObject"; };

function Scratch::registerFunctionHelp(%funcName, %arguments, %category, %desc, %argCount) {
	// %arguments - Argument names. Example: "myArg anotherArg yetAnotherArg" (Delimited by spaces).
	// %category  - Category the function is in. Example: "String Functions"
	// %desc      - Description of the function.
	// %argCount  - How many args are used in this function.
	%this = ScratchFuncHelp;
	for(%i=0;%i<%this.size();%i++) {
		%c_data = %this.at(%i);
		%c_name = getField(%c_data, 0);
		%c_args = getField(%c_data, 1);
		%c_cat  = getField(%c_data, 2);
		%c_desc = getField(%c_data, 3);
		if(%c_name !$= %funcName) continue;
		echo("Scratch::registerFunctionHelp() - Function with name " @ %c_name @ " is already documented; overwriting...");
		%this.setAt(%i, %c_name TAB %arguments TAB %category TAB %desc TAB %argCount);
		return 1;
	}
	%this.push_back(%funcName TAB %arguments TAB %category TAB %desc TAB %argCount);
	%this.cat[getSafeVariableName(%category)] = %this.cat[getSafeVariableName(%category)] SPC %this.size()-1;
	for(%i=0;%i<=%this.catCount;%i++) {
		if(%i == %this.catCount) { %this.catTable[-1+%this.catCount++] = %category; break; }
		if(%this.catTable[%i] !$= %category) continue;
		break;
	}
}

function Scratch::getFunctionHelpStr(%funcName) {
	%this = ScratchFuncHelp;
	for(%i=0;%i<=%this.size();%i++) {
		if(%i == %this.size()) { error("ERROR: Scratch::getFunctionHelpStr() - Cannot find documentation for function '" @ %funcName @ "'."); return ""; }
		%c_data = %this.at(%i);
		%c_name = getField(%c_data, 0);
		%c_args = getField(%c_data, 1);
		%c_cat  = getField(%c_data, 2);
		%c_desc = getField(%c_data, 3);
		%c_argc = getField(%c_data, 4);
		if(%c_name !$= %funcName) continue;
		break;
	}
	for(%i=0;%i<getWordCount(%c_args);%i++) %funcStr = (%funcStr $= "" ? "" : %funcStr @ ", ") @ "%" @ getWord(%c_args, %i);
	return "<font:Consolas:14>" @ %c_name @ "(" @ %funcStr @ ")<br><font:Consolas:12><color:888888>" @ %c_cat @ "<color:000000><font:Consolas:14><br><br>" @ %c_desc;
}

// Default functions

// String
Scratch::registerFunctionHelp("strLen",              "str",                                "String Manipulation", "Returns the length of %str.", 1);
Scratch::registerFunctionHelp("strPos",              "sourceStr searchFor start",          "String Manipulation", "Returns the position of %searchFor in %sourceStr. Returns -1 if it cannot find the string.", 3);
Scratch::registerFunctionHelp("strIPos",             "sourceStr searchFor start",          "String Manipulation", "A case-insensitive version of strPos", 3);
Scratch::registerFunctionHelp("getSubStr",           "sourceStr start length",             "String Manipulation", "Get a portion of %sourceStr, starting at %start and being %length characters long.", 3);
Scratch::registerFunctionHelp("strCmp",              "stringOne stringTwo",                "String Manipulation", "Compare two strings. Returns 0 if the strings match, and -1 if they don't.", 2);
Scratch::registerFunctionHelp("strICmp",             "stringOne stringTwo",                "String Manipulation", "A case-insensitive version of strCmp", 2);
Scratch::registerFunctionHelp("strChr",              "sourceStr startAtString",            "String Manipulation", "If %startAtString is found in %sourceStr, then it returns everything after where it found the string.", 2);
Scratch::registerFunctionHelp("stripChars",          "sourceStr string",                   "String Manipulation", "Strips all instances of every character in %string from %sourceStr.", 2);
Scratch::registerFunctionHelp("strReplace",          "sourceStr searchStr replaceWithStr", "String Manipulation", "Replace all instances of %searchStr in %sourceStr with %replaceWithStr.", 3);
Scratch::registerFunctionHelp("strIReplace",         "sourceStr searchStr replaceWithStr", "String Manipulation", "A case-insensitive version of strReplace", 3);
Scratch::registerFunctionHelp("stripMLControlChars", "str",                                "String Manipulation", "Strips all ML Control Chars (e.g. < font:Verdana Bold:16 >) from %str.", 1);

// Words
Scratch::registerFunctionHelp("firstWord",    "str",               "String Manipulation", "Gets the first word in %str.", 1);
Scratch::registerFunctionHelp("getWord",      "str index",         "String Manipulation", "Gets the word at %index in %str.", 2);
Scratch::registerFunctionHelp("getWordCount", "str",               "String Manipulation", "Gets the amount of words in %str.", 1);
Scratch::registerFunctionHelp("getWords",     "str start end",     "String Manipulation", "Gets the words between index %start and index %end in %str.", 3);
Scratch::registerFunctionHelp("removeWord",   "str index",         "String Manipulation", "Remove the word at %index in %str.", 2);
Scratch::registerFunctionHelp("restWords",    "str",               "String Manipulation", "Gets all words after the first word in %str.", 1);
Scratch::registerFunctionHelp("setWord",      "str index newWord", "String Manipulation", "Set the word at %index to %newWord in %str.", 3);

// Fields
Scratch::registerFunctionHelp("getField",      "str index",          "String Manipulation", "Get the field at %index in %str.", 2);
Scratch::registerFunctionHelp("getFieldCount", "str",                "String Manipulation", "Get the amount of fields in %str.", 1);
Scratch::registerFunctionHelp("getFields",     "str start end",      "String Manipulation", "Get the fields between index %start and index %end in %str.", 3);
Scratch::registerFunctionHelp("removeField",   "str index",          "String Manipulation", "Remove the field at %index in %str.", 2);
Scratch::registerFunctionHelp("setField",      "str index newField", "String Manipulation", "Set the field at %index to %newField in %str.", 3);

// Records
Scratch::registerFunctionHelp("getRecord",      "str index",           "String Manipulation", "Get the record at %index in %str.", 2);
Scratch::registerFunctionHelp("getRecordCount", "str",                 "String Manipulation", "Get the amount of records in %str.", 1);
Scratch::registerFunctionHelp("getRecords",     "str start end",       "String Manipulation", "Get the records between index %start and index %end in %str.", 3);
Scratch::registerFunctionHelp("removeRecord",   "str index",           "String Manipulation", "Remove the record at %index in %str.", 2);
Scratch::registerFunctionHelp("setRecord",      "str index newRecord", "String Manipulation", "Set the record at %index to %newRecord in %str.", 3);

// Scripting
Scratch::registerFunctionHelp("eval", "string",   "Scripts", "Evaluate torquescript. Returns the result.", 1);
Scratch::registerFunctionHelp("exec", "fileName", "Scripts", "Execute a .cs or .gui file.", 1);

// Debugging
Scratch::registerFunctionHelp("echo",      "str0 str1 str[n]...", "Debugging", "Prints the string into the console & logs it in console.log", 1);
Scratch::registerFunctionHelp("warn",      "str0 str1 str[n]...", "Debugging", "Similar to echo, but it prints gray text", 1);
Scratch::registerFunctionHelp("error",     "str0 str1 str[n]...", "Debugging", "Similar to echo, but it prints red text and provides a backtrace for debugging", 1);
Scratch::registerFunctionHelp("backTrace", "",                    "Debugging", "Echos a string in console that details what functions were called to get to this function", 0);
Scratch::registerFunctionHelp("trace",     "boolean",             "Debugging", "", 1);

// Server-side
Scratch::registerFunctionHelp("findClientByBL_ID", "bl_id", "Server Miscellaneous", "Find a client by their Blockland ID. Returns -1 if not found.", 1);
Scratch::registerFunctionHelp("findClientByName",  "name",  "Server Miscellaneous", "Find a client by their player name. Returns -1 if not found.", 1);
Scratch::registerFunctionHelp("announce",          "str",   "Server Miscellaneous", "Displays %str to everyone in the server.", 1);
Scratch::registerFunctionHelp("talk",              "str",   "Server Miscellaneous", "Displays %str to everyone in the server in the format of \"CONSOLE: %str\"", 1);

// Math
Scratch::registerFunctionHelp("mFloatLength", "num numDecimals", "Math", "Limit %num to %numDecimals amount of decimals after '.'", 2);
Scratch::registerFunctionHelp("mFloor",       "num",             "Math", "Forces %num to the whole number if decimals are present.", 1);
Scratch::registerFunctionHelp("mClamp",       "num min max",     "Math", "Clamps %num between %min and %max.", 3);
Scratch::registerFunctionHelp("mCeil",        "num",             "Math", "Forces %num to the next whole number if decimals are present.", 1);
Scratch::registerFunctionHelp("mSqrt",        "num",             "Math", "Calculate the square root of %num.", 1);
Scratch::registerFunctionHelp("mAbs",         "num",             "Math", "Returns the absolute value of %num.", 1);
Scratch::registerFunctionHelp("mPow",         "num p",           "Math", "Returns %num raised to the %p'th power.", 2);

// Vectors
Scratch::registerFunctionHelp("vectorDist",       "vectorA vectorB", "Vector Math", "Calculate the distance between %vectorA and %vectorB.", 2);
Scratch::registerFunctionHelp("vectorAdd",        "vectorA vectorB", "Vector Math", "Add %vectorB to %vectorA.", 2);
Scratch::registerFunctionHelp("vectorSub",        "vectorA vectorB", "Vector Math", "Subtract %vectorB from %vectorA.", 2);
Scratch::registerFunctionHelp("vectorScale",      "vector num",      "Vector Math", "Scale %vector by %num.", 2);
Scratch::registerFunctionHelp("vectorLen",        "vector",          "Vector Math", "Calculate the length of %vector.", 1);
Scratch::registerFunctionHelp("vectorNormalize",  "vector",          "Vector Math", "Returns a scaled down vector, such that the length of %vector is 1.", 1);
Scratch::registerFunctionHelp("vectorDot",        "vectorA vectorB", "Vector Math", "Calculate the dot product of %vectorA and %vectorB.", 2);
Scratch::registerFunctionHelp("vectorCross",      "vectorA vectorB", "Vector Math", "Calculate the cross product of %vectorA and %vectorB.", 2);
Scratch::registerFunctionHelp("getBoxCenter",     "box",             "Vector Math", "Get the center point of a box.", 1);

// Trigonometry
Scratch::registerFunctionHelp("mAcos", "num",            "Trigonometry Math", "Return the arc-cosine of %num, in radians.", 1);
Scratch::registerFunctionHelp("mAsin", "num",            "Trigonometry Math", "Return the arc-sine of %num, in radians.", 1);
Scratch::registerFunctionHelp("mAtan", "numRise numRun", "Trigonometry Math", "Return the slope of a line with the given %numRise and %numRun.", 2);
Scratch::registerFunctionHelp("mCos",  "num",            "Trigonometry Math", "Return the cosine of %num, in radians.", 1);
Scratch::registerFunctionHelp("mSin",  "num",            "Trigonometry Math", "Return the sine of %num, in radians.", 1);
Scratch::registerFunctionHelp("mTan",  "num",            "Trigonometry Math", "Return the tangent of %num, in radians.", 1);

// Objects
Scratch::registerFunctionHelp("isObject", "object",     "Objects", "Check to see if any object has an ID / name of %object. Returns 1 if so, and 0 if not.", 1);
Scratch::registerFunctionHelp("nameToID", "objectName", "Objects", "If an object exists with the name %objectName, then it returns that object's ID.", 1);

// Schedules / Timing
Scratch::registerFunctionHelp("getSimTime",       "",                                 "Schedules / Timing", "Get the amount of time, in milliseconds, since Blockland was opened.", 0);
Scratch::registerFunctionHelp("getRealTime",      "",                                 "Schedules / Timing", "Get the amount of time, in milliseconds, since the PC was booted up.", 0);
Scratch::registerFunctionHelp("schedule",         "time refobject command arguments", "Schedules / Timing", "Schedules a function call to %command with the arguments %arguments %time milliseconds in the future. Returns a ScheduleID.", 4);
Scratch::registerFunctionHelp("isEventPending",   "scheduleID",                       "Schedules / Timing", "Returns 1 if a schedule by the ID %scheduleID has expired yet.", 1);
Scratch::registerFunctionHelp("getTimeRemaining", "scheduleID",                       "Schedules / Timing", "Returns the amount of time the schedule that goes by the ID %scheduleID has before it finishes.", 1);
Scratch::registerFunctionHelp("cancel",           "scheduleID",                       "Schedules / Timing", "Cancels the schedule that goes by the ID %scheduleID.", 1);

// Files
Scratch::registerFunctionHelp("fileDelete", "file",                 "File Manipulation", "Permanently delete %file if it exists.", 1);
Scratch::registerFunctionHelp("fileCopy",   "file fileDestination", "File Manipulation", "Copy %file to %fileDestination. Does not delete %source after it's done copying.", 1);
Scratch::registerFunctionHelp("isFile",     "file",                 "File Manipulation", "Checks if %file exists.", 1);

// Random
Scratch::registerFunctionHelp("getRandom",     "min max", "Random", "Gets a number between %min and %max.", 2);
Scratch::registerFunctionHelp("setRandomSeed", "num",     "Random", "Sets the getRandom() seed to %num.", 1);
Scratch::registerFunctionHelp("getRandomSeed", "num",     "Random", "Get the current getRandom() seed.", 1);