
if(!isObject(ScratchHighlightContainer)) new SimSet("ScratchHighlightContainer");

$CACHE::Scratch::HighlightSystem::Outline = 2;

function Scratch::createHighlightObject(%btmConnector) {
	%newObj = new GuiSwatchCtrl() {
		profile = "GuiScratchNoModalProfile";
		horizSizing = "right";
		vertSizing = "bottom";
		position = "0 0";
		extent = "174 27";
		minExtent = "8 2";
		visible = "1";
		color = "0 0 0 0";

		new GuiBitmapBorderCtrl() {
				profile = "ScratchBlockOutlineProfile";
				horizSizing = "width";
				vertSizing = "height";
				position = "0 0";
				extent = "174 20";
				minExtent = "8 2";
				visible = "1";
		};
	};
	if(%btmConnector) {
		%btmObj = new GuiBitmapBorderCtrl() {
				profile = "ScratchBlockOutlineProfile";
				horizSizing = "right";
				vertSizing = "top";
				position = "10 7";
				extent = "19 20";
				minExtent = "8 2";
				visible = "1";
		};
		%newObj.add(%btmObj);
	}
	ScratchHighlightContainer.add(%newObj);
	return %newObj;
}

function Scratch::highlightElement(%this, %type, %link) {
	if(!isObject(%this)) return;
	if(%link && %type != 0) %link = 0;
	if(%link == 1) %this = Scratch::getHighestTopConnection(%this);
	if(%this.container) Scratch::highlightContainer(%this, %type, %link);
	else Scratch::highlightBlock(%this, %type, %link);
}

function Scratch::highlightBlock(%this, %type, %link) {
	%olSize = $CACHE::Scratch::HighlightSystem::Outline; // Outline size
	%blSize = %olSize;
	%bottom = (%this.bottom && !isObject(%this.bottomConnection) ? 1 : 0);

	if(%type == 1) %bottom = 1;	

	if(!isObject(%newObj = %this.highlightObj)) {
		%newObj            = Scratch::createHighlightObject(%bottom);
		%this.highlightObj = %newObj;
		Scratch_Content.add(%newObj);
	}
	if(%type >= 3 || %type < 0) {
		warn("WARNING: Scratch::highlightBlock() - Invalid highlight type \"" @ %type @ "\"");
		warn("WARNING: Scratch::highlightBlock() - Types:");
		warn("WARNING: Scratch::highlightBlock() -    > 0 : Entire block");
		warn("WARNING: Scratch::highlightBlock() -    > 1 : Top of block");
		warn("WARNING: Scratch::highlightBlock() -    > 2 : Bottom of block");
		%type = 0;
	}
	switch(%type) {
		case 0: // entire block
			%pos = %this.getLeft()-%blSize SPC %this.getTop()-%blSize;
			%ext = %this.getWidth()+(%blSize*2) SPC %this.getHeight()+(%blSize*2);
		case 1: // top
			%ext    = %this.getWidth() SPC %this.getHeight();
			%pos    = %this.getLeft() SPC %this.getTop()-(getWord(%ext, 1)-7);
			if(%bottom) %newObj.getObject(1).resize(-20, -20, 8, 2);
			%bottom = 0;
		case 2: // bottom
			%pos = %this.getLeft() SPC %this.getTop();
			%ext = %this.getWidth() SPC %this.getHeight()+%blSize;
	}
	%newObj.resize(getWord(%pos, 0), getWord(%pos, 1), getWord(%ext, 0), getWord(%ext, 1));
	if(%bottom) {
		%offsetX = mAbs(%newObj.getLeft()-%this.getLeft());
		%offsetY = mAbs(%newObj.getTop()-%this.getTop());
		%objBtm  = %this.bottomConnector;
		%btmCon  = %newObj.getObject(1);
		%pos     = %objBtm.getLeft()-%olSize SPC %objBtm.getTop()-%olSize;
		%ext     = %objBtm.getWidth()+(%olSize*%olSize) SPC %objBtm.getHeight()+(%olSize*%olSize);
		%btmCon.resize(getWord(%pos, 0)+%offsetX, getWord(%pos, 1)+%offsetY, getWord(%ext, 0), getWord(%ext, 1));
	}
	Scratch_Content.bringToFront(%newObj);
	%this.highlighted = 1;
	if(%link && isObject(%obj = %this.bottomConnection)) {
		Scratch::highlightBlock(%obj, %type, 2);
	}
}

function Scratch::highlightContainer(%this, %type, %link) {
	%olSize = $CACHE::Scratch::HighlightSystem::Outline; // Outline size
	%blSize = %olSize;
	%bottom = (%this.bottom && !isObject(%this.bottomConnection) ? 1 : 0);
	if(!isObject(%newObj = %this.highlightObj)) {
		%newObj            = Scratch::createHighlightObject(%bottom);
		%this.highlightObj = %newObj;
		Scratch_Content.add(%newObj);
		Scratch_Content.bringToFront(%newObj);
	}
	if(%type >= 3 || %type < 0) {
		warn("WARNING: Scratch::highlightBlock() - Invalid highlight type \"" @ %type @ "\"");
		warn("WARNING: Scratch::highlightBlock() - Types:");
		warn("WARNING: Scratch::highlightBlock() -    > 0 : Entire block");
		warn("WARNING: Scratch::highlightBlock() -    > 1 : Top of block");
		warn("WARNING: Scratch::highlightBlock() -    > 2 : Bottom of block");
		%type = 0;
	}
	switch(%type) {
		case 0: // entire block
			%pos = %this.getLeft()-%blSize SPC %this.getTop()-%blSize;
			%ext = %this.barConnector.getWidth()+(%blSize*2) SPC %this.getHeight()+(%blSize*2);
			if(!isObject(%newObj0 = %this.container0.highlightObj)) {
				%newObj0                      = Scratch::createHighlightObject(%bottom);
				%this.container0.highlightObj = %newObj0;
				Scratch_Content.add(%newObj0);
			}
			if(!isObject(%newObj1 = %this.container1.highlightObj)) {
				%newObj1                      = Scratch::createHighlightObject(%bottom);
				%this.container1.highlightObj = %newObj1;
				Scratch_Content.add(%newObj1);
			}
		case 1: // top
			%pos    = %this.getLeft() SPC %this.getTop()-%blSize;
			%ext    = %this.getWidth() SPC %this.getHeight();
			if(%bottom) %newObj.getObject(1).resize(-20, -20, 8, 2);
			%bottom = 0;
		case 2: // bottom
			%pos = %this.getLeft() SPC %this.getTop();
			%ext = %this.getWidth() SPC %this.getHeight()+%blSize;
	}
	%newObj.resize(getWord(%pos, 0), getWord(%pos, 1), getWord(%ext, 0), getWord(%ext, 1));
	if(%type == 0) {
		// Top container
		%pos = (%this.getLeft()+%this.container0.getLeft())-%blSize SPC (%this.getTop()+%this.container0.getTop())-%blSize;
		%ext = %this.container0.getWidth()+(%blSize*2) SPC %this.container0.getHeight()+(%blSize*2);
		%newObj0.resize(getWord(%pos, 0), getWord(%pos, 1), getWord(%ext, 0), getWord(%ext, 1));

		// Bottom container
		%pos = (%this.getLeft()+%this.container1.getLeft())-%blSize SPC ((%this.getTop()+%this.container1.getTop())-%blSize)+9;
		%ext = %this.container1.getWidth()+(%blSize*2) SPC (%this.container1.getHeight()+(%blSize*2))-9;
		%newObj1.resize(getWord(%pos, 0), getWord(%pos, 1), getWord(%ext, 0), getWord(%ext, 1));

		// Top container bottom connector
		%objBtm  = %this.container0.bottomConnector;
		%btmCon  = %newObj0.getObject(1);
		%pos     = %objBtm.getLeft()-%olSize SPC %objBtm.getTop()-%olSize;
		%ext     = %objBtm.getWidth()+(%olSize*%olSize) SPC %objBtm.getHeight()+(%olSize*%olSize);
		%btmCon.resize(getWord(%pos, 0), getWord(%pos, 1), getWord(%ext, 0), getWord(%ext, 1));

		// Bottom container bottom connector
		%objBtm = %this.bottomConnector;
		%btmCon = %newObj1.getObject(1);
		%pos    = %objBtm.getLeft()-%olSize SPC %objBtm.getTop()-%olSize;
		%ext    = %objBtm.getWidth()+(%olSize*%olSize) SPC %objBtm.getHeight()+(%olSize*%olSize);
		%btmCon.resize(getWord(%pos, 0), getWord(%pos, 1), getWord(%ext, 0), getWord(%ext, 1));
	}
	if(%bottom) {
		%offsetX = mAbs(%newObj.getLeft()-%this.getLeft());
		%offsetY = mAbs(%newObj.getTop()-%this.getTop());
		%objBtm  = %this.bottomConnector;
		%btmCon  = %newObj.getObject(1);
		%pos     = %objBtm.getLeft()-%olSize SPC %objBtm.getTop()-%olSize;
		%ext     = %objBtm.getWidth()+(%olSize*%olSize) SPC %objBtm.getHeight()+(%olSize*%olSize);
		%btmCon.resize(getWord(%pos, 0)+%offsetX, getWord(%pos, 1)+%offsetY, getWord(%ext, 0), getWord(%ext, 1));
	}
	Scratch_Content.bringToFront(%newObj);
	if(%type == 0) {
		Scratch_Content.bringToFront(%newObj1);
		Scratch_Content.bringToFront(%newObj0);
	}
	%this.highlighted = 1;
	if(%link && isObject(%obj = %this.bottomConnection)) Scratch::highlightBlock(%obj, %type, 2);
}

function Scratch::clearBlockHighlight(%this, %link) {
	if(!isObject(%this)) return;
	if(!isObject(%obj = %this.highlightObj)) return;
	if(%this.container) {
		%this.container0.highlightObj.delete();
		%this.container1.highlightObj.delete();
	}
	%obj.delete();
	%this.highlighted = 0;
	if(%link && isObject(%obj = %this.bottomConnection)) Scratch::clearBlockHighlight(%obj, 1);
}