if(!isObject(ScratchCode)) {
	new ScriptObject(ScratchCode) {
		class = "ArrayObject";
	};
}
function ScratchCode::addCode(%this, %actionName, %args, %line) {
	if(%line+1 >= Scratch_LineButtons.getCount()) {
		ScratchGui.createLineButtonsToIndex(%line+2);
	}
	%this.setAt(%line, %actionName TAB %args);
}
function isNumber(%num) {
	return (atof(%num) $= %num);
}
function getSafeArgument(%str) {
	if(stripChars(%str, "0123456789.-") $= "" && isNumber(%str)) {
		return %str;
	}
	if(getSubStr(%str, 0, 2) $= "0x" && stripChars(%str, "0x123456789ABCDEF") $= "") {
		// Hedidecimal
		return %str;
	}
	%str = expandEscape(%str);
	%str = strReplace(%str, "'", "\\'");
	%str = strReplace(%str, "\"", "\\\"");
	%str = collapseEscape(%str);
	%str = "\"" @ %str @ "\"";
	return %str;
}
function buildArguments(%str) {
	for(%i=0;%i<getFieldCount(%str);%i++) {
		%ret = (%ret $= "" ? "" : %ret @ ", ") @ getSafeArgument(getField(%str, %i));
	}
	return %ret;
}
function ScratchCode::scratchToEval(%this) {
	%indent = 0;
	%SO = new FileObject();
	%SO.openForWrite("Add-Ons/Client_Scratch/EvalTest.cs");
	for(%i=0;%i<%this.size();%i++) {
		%actName     = getField(%this.at(%i), 0);
		%args        = removeField(%this.at(%i), 0);
		%toAdd       = "";
		%indent      = "";
		%regenIndent = 0;
		for(%j=0;%j<%indent;%j++) %indent = %indent @ "\t";
		switch$(%actName) {
			case "IF":        %indent++; %regenIndent = 0; %toAdd = "if(" @ %args @ ") {";
			case "IF_ELIF":   %toAdd = "} else if(" @ %args @ ") {";
			case "IF_ELSE":   %toAdd = "} else {";
			case "IF_END":    %indent--; %regenIndent = 1; %toAdd = "}";
			case "OBJ":       %indent++; %regenIndent = 0; %toAdd = "new " @ getField(%args, 0) @ "(" @ buildArguments(removeField(%args, 0)) @ ") {";
			case "OBJ_END":   %indent--; %regenIndent = 1; %toAdd = "};";
			case "WHILE":     %indent++; %regenIndent = 0; %toAdd = "while(" @ %args @ ") {";
			case "WHILE_END": %indent--; %regenIndent = 1; %toAdd = "}";
			case "FOR":
				// [String:Variable Name]\t[String:Initial Value]\t[String:Statement]\t[String:Step]
				%indent++;
				%var   = getField(%args, 0);
				%toAdd = "for(" @ %var @ "=" @ getSafeVariable(getField(%args, 1)) @ ";" @ getField(%args, 2) @ ";" @ getField(%args, 3) @ ") {";
			case "FOR_END":
				%indent--; %regenIndent = 1;
				%toAdd = "}";
			case "SET_VAR":
				// [0/1:Local or Global]\t[Arg Name]\t[String:Array Value]\t[0/1/2/3/4/5/6/7:Operator]\t[String:Value]
				switch(getField(%args, 3)) {
					case 0: %op = "=";
					case 1: %op = "+=";
					case 2: %op = "-=";
					case 3: %op = "*=";
					case 4: %op = "/=";
					case 5: %op = "%=";
					case 6: %op = "&=";
					case 7: %op = "^=";
				}
				%toAdd = (getField(%args, 0) $= "0" ? "%" : "$") @ getField(%args, 1) @ (getField(%args, 2) $= "" ? "" : "[" @ getField(%args, 2) @ "]") SPC %op SPC getField(%args, 4) @ ";";
			case "FUNC":
				// [String:Function Name]\t[String:Argument names delimited by space]
				for(%j=0;%j<getWordCount(getField(%args, 1));%j++) %funcArgs = (%funcArgs $= "" ? "" : %funcArgs @ ", %") @ getWord(getField(%args, 1), %j);
				%indent++;
				%toAdd = "function " @ getSafeVariableName(getField(%args, 0)) @ "(" @ %funcArgs @ ") {";
			case "FUNC_END":
				%indent--;
				%regenIndent = 1;
				%toAdd = "}";
			case "CALLFUNC":
				%toAdd = getSafeVariableName(getField(%args, 0)) @ "(" @ buildArguments(removeField(%args, 0)) @ ");";
			default:
				warn("WARNING: scratchToEval() - Unimplemented op code '" @ %actName @ "'"); continue;
		}
		if(%regenIndent) { %indent  = ""; for(%j=0;%j<%indent;%j++) %indent = %indent @ "\t"; }
		%SO.writeLine(%indent @ %toAdd);
	}
	%SO.close();
	%SO.delete();
	echo("\c4Done generating script");
}