$Scratch::NetVersion = "0.1.1";

package ScratchNet_ClientScoping {
	function Scratch::duplicateBlock(%this, %bypass) {
		if(!ScratchClientTCP.connected || %bypass) { return Parent::duplicateBlock(%this); }
		ScratchClientTCP.sendData("BLOCK_DUPLICATE" SPC %this.ghostID);
	}
	function SpawnMenuButtonHandler::onMouseDown(%this) {
		if(!ScratchClientTCP.connected) { Parent::onMouseDown(%this); return; }
		%group = %this.getGroup();
		%data  = %group.blockType;
		%data  = %data TAB %group.blockArgs;
		Scratch::toggleProjectManager();
		ScratchClientTCP.sendData("BLOCK_SPAWN" SPC %data);
	}
	function Scratch::deleteBlock(%this, %byPass) {
		if(!ScratchClientTCP.connected) return Parent::deleteBlock(%this);
		if(!%byPass) {
			ScratchClientTCP.sendData("BLOCK_DELETE" SPC %this.ghostID);
		} else Parent::deleteBlock(%this);
	}
	function Scratch::handleDragStatus(%this) {
		if(!ScratchClientTCP.connected) { Parent::handleDragStatus(%this); return; }
		if(!%this.status $= "DRAG") return;
		Parent::handleDragStatus(%this);
		ScratchClientTCP.sendData("BLOCK_MOVE" SPC %this.ghostID SPC %this.getPosition());
	}
	function Scratch::attachBlock(%this, %other, %side, %bypass) {
		if(!ScratchClientTCP.connected) return Parent::attachBlock(%this, %other, %side);
		if(%bypass) return Parent::attachBlock(%this, %other, %side);
		if(%other.argType !$= "") {
			%group = %other.getGroup();
			%found = -1;
			for(%i=0;%i<%group.blockArgCount;%i++) {
				if(%group.blockArgObject[%i].getId() != %other.getId()) continue;
				%found = %i;
				break;
			}
			if(%found == -1) return;
			ScratchClientTCP.sendData("BLOCK_ATTACH_INPUTBLOCK" SPC %this.ghostID SPC %group.ghostID SPC %found);
			return 1;
		}
		ScratchClientTCP.sendData("BLOCK_ATTACH" SPC %this.ghostID SPC %other.ghostID SPC %side);
		return 1;
	}
	function Scratch::detachBlock(%this, %side, %bypass) {
		if(!ScratchClientTCP.connected) return Parent::detachBlock(%this, %side);
		if(%bypass) return Parent::detachBlock(%this, %side);
		ScratchClientTCP.sendData("BLOCK_DETACH" SPC %this.ghostID SPC %side);
		return Parent::detachBlock(%this, %side);
	}
	function Scratch::setTextBoxWidth(%textBox, %container, %bypass) {
		if(!ScratchClientTCP.connected) { Parent::setTextBoxWidth(%textBox, %container); return; }
		Parent::setTextBoxWidth(%textBox, %container);
		if(!%bypass) {
			%group = %container.getGroup();
			%found = -1;
			for(%i=0;%i<%group.blockArgCount;%i++) {
				if(%group.blockArgObject[%i].getId() != %container.getId()) continue;
				%found = %i;
				break;
			}
			if(%found == -1) return;
			cancel($CACHE::Scratch::Net_Client::TextTimeout[%textBox]);
			$CACHE::Scratch::Net_Client::TextTimeout[%textBox] = ScratchClientTCP.schedule(300, sendData, "BLOCK_SETTEXT" SPC %container.getGroup().ghostID SPC %found SPC %textBox.getValue());
		}
	}
	function Scratch::addDelimiter(%this, %type, %cannotDelete, %bypass) {
		if(!ScratchClientTCP.connected) { Parent::addDelimiter(%this, %type, %cannotDelete); return; }
		if(%bypass || %cannotDelete) Parent::addDelimiter(%this, %type, %cannotDelete);
		else ScratchClientTCP.sendData("BLOCK_ADD_DELIMITER" SPC %this.ghostID SPC %type);
	}
	function Scratch::removeDelimiter(%this, %bypass) {
		if(!ScratchClientTCP.connected) { Parent::removeDelimiter(%this); return; }
		if(%bypass) Parent::removeDelimiter(%this);
		else ScratchClientTCP.sendData("BLOCK_REMOVE_DELIMITER" SPC %this.ghostID);
	}
	function ScratchGui::ClearCode(%this, %confirm, %bypass) {
		if(!%confirm || %bypass || !ScratchClientTCP.connected) { Parent::ClearCode(%this, %confirm); return; }
		ScratchClientTCP.sendData("CLEAR_CODE");
	}
};

if(!isObject(ScratchClientTCP)) new TCPObject(ScratchClientTCP);
if(!isObject(ScratchMouseGroup)) new SimSet(ScratchMouseGroup);

function Scratch::connectToServer(%ip, %port) {
	if(!$CACHE::Scratch::Net_Server::Hosting) ScratchGui.ClearCode(1);
	if(ScratchClientTCP.connected) {
		ScratchClientTCP.sendData("DISCONNECT");
		ScratchClientTCP.disconnect();
	}
	ScratchClientTCP.ip         = %ip @ ":" @ %port;
	ScratchClientTCP.retryCount = 0;
	ScratchClientTCP.connect(%ip @ ":" @ %port);
	Scratch::collabConnStatus("Connecting to " @ %ip @ ":" @ %port @ "...", 0);
	echo("Scratch Client " @ $Scratch::NetVersion @ " | Attempting to connect to " @ %ip @ ":" @ %port);
}

function Scratch::clientDisconnect() {
	ScratchClientTCP.disconnect();
	ScratchClientTCP.onDisconnect();
	if($CACHE::Scratch::Net_Server::Hosting) {
		for(%i=0;%i<ScratchServerClientGroup.getCount();%i++) {
			%object = ScratchServerClientGroup.getObject(%i);
			%object.sendData("DISCONNECT");
			%object.disconnect();
			%object.schedule(1, delete);
		}
		ScratchServerTCP.delete();
		new TCPObject(ScratchServerTCP);
		$CACHE::Scratch::Net_Server::Hosting = 0;
		
	}
}

function ScratchClientTCP::onConnectFailed(%this) {
	if(%this.retryCount++ >= 4) {
		Scratch::collabConnStatus("Connection failure, retrying (attempt # " @ %this.retryCount @ ")...", 1);
		%this.schedule(1, connect, %this.ip);
	} else {
		Scratch::collabConnStatus("Connection failure", 1);
	}
}

function ScratchClientTCP::onDNSFailed(%this) {
	Scratch::collabConnStatus("DNS lookup failure (?)", 1);
}

function ScratchClientTCP::onConnected(%this) {
	Scratch::collabConnStatus("Waiting for authentication end...");
	activatePackage(ScratchNet_ClientScoping);
	Scratch::openCollabMenu();
	%this.connected = 1;
	echo("Scratch Client " @ $Scratch::NetVersion @ " | Connected, waiting for status");
	%this.sendData("INIT");
}

function ScratchClientTCP::onDisconnect(%this) {
	for(%i=0;%i<ScratchMouseGroup.getCount();%i++) {
		%object = ScratchMouseGroup.getObject(%i);
		%object.textObj.delete();
		%object.delete();
	}
	Scratch::CollabMenu_clear();
	deleteVariables("$CACHE::Scratch::Net_Client::ScopedObjForID*");
	if(%this.authed) messageBoxYesNo("Confirmation - Scratch Gui", "Do you want to clear the workspace?", "ScratchGui.ClearCode(1);");
	echo("Scratch Client " @ $Scratch::NetVersion @ " | Disconnected from server");
	%this.connected = 0;
	%this.authed = 0;
}

function ScratchClientTCP::sendData(%this, %line) {
	%this.send(%line @ "\r\n");
}

function Scratch::getScopedObjectByGhostID(%ghostID, %blockType, %blockArgs) {
	if(!isObject($CACHE::Scratch::Net_Client::ScopedObjForID[%ghostID])) {
		if(%blockType $= "" && %blockArgs $= "") return 0;
		%newObj         = Scratch::createPremadeBlock(%blockType, %blockArgs);
		Scratch_Content.add(%newObj);
		%newObj.ghostID = %ghostID;
		$CACHE::Scratch::Net_Client::ScopedObjForID[%ghostID] = %newObj;
	}
	return $CACHE::Scratch::Net_Client::ScopedObjForID[%ghostID];
}

function Scratch::getScopedMouse(%name, %doNotCreate) {
	%ClientID = getStringCRC(%name);
	if(!isObject(%obj = $CACHE::Scratch::Net_Client::ScopedMouse[%ClientID])) {
		if(%doNotCreate) return 0;
		%newObj = new GuiBitmapCtrl() {
			extent = "16 32";
			profile = "GuiScratchNoModalProfile";
			bitmap = "base/client/ui/CUR_3darrow";
		};
		%obj = %newObj;
		%mlTextObj = new GuiMLTextCtrl() {
			extent = "150 16";
			profile = "GuiScratchNoModalProfile";
			text = "<font:Franklin Gothic Demi:14>" @ %name;
		};
		%newObj.textObj = %mlTextObj;
		ScratchGui.add(%obj);
		ScratchGui.add(%obj.textObj);
		ScratchMouseGroup.add(%obj);
		$CACHE::Scratch::Net_Client::ScopedMouse[%ClientID] = %newObj.getId();
	}
	return $CACHE::Scratch::Net_Client::ScopedMouse[%ClientID];
}

function ScratchClientTCP::onLine(%this, %line) {
	%act  = firstWord(%line);
	%line = removeWord(%line, 0);
	if(!%this.authed) {
		switch$(%act) {
			case "STATUS_OK":
				Scratch_CollabConnect.setVisible(0);
				Scratch::openCollabMenu();
				%netName = %line;
				echo("Scratch Client " @ $Scratch::NetVersion @ " | Done sending client info!");
				%this.authed = 1;
				if($CACHE::Scratch::Net_Server::Hosting) %this.sendData("HOST_ATTEMPT" SPC $CACHE::Scratch::Net_Server::HostPassword);
				%this.sendData("REQUEST_GHOST_INFO");
				ScratchNet::sendMouseInfo();
				CloseMessagePopup();
			case "STATUS_PASS":
				%this.sendData("PASS" SPC $CACHE::Scratch::Net_Client::JoinPassword);
			case "STATUS_INVALID":
				%this.disconnect();
				%this.onDisconnect();
				echo("Scratch Client " @ $Scratch::NetVersion @ " | Incorrect password; disconnected");
				Scratch::collabConnStatus("Incorrect password", 1);
			case "STATUS_NAME":
				%this.sendData("NAME " @ $Pref::Player::NetName);
		}
		return;
	}
	switch$(%act) {
		case "CLIENTCMD":
			%funcName = "scratchClientCmd" @ getSafeVariableName(getField(%line, 0));
			for(%i=1;%i<getFieldCount(%line);%i++) %args = (%args $= "" ? "" : %args @ ", ") @ getSafeArgument(getField(%line, %i));
			if(!isFunction(%funcName)) return;
			eval(%funcName @ "(" @ %args @ ");");
		case "PING": %this.sendData("PONG");
		case "CLEAR_CODE": ScratchGui.clearCode(1, 1);
		case "BLOCK_ERROR":
			if(isObject($CACHE::Scratch::BlockSystem::LastDraggedMouseEvent)) {
				$CACHE::Scratch::BlockSystem::LastDraggedMouseEvent.onMouseUp();
			}
			messageBoxOk("Error - Scratch Editor", getWords(%line, 0, 500));
		case "BLOCK_DELETE": if($CACHE::Scratch::Net_Server::Hosting) return;
			%ghostID  = getWord(%line, 0);
			%object   = Scratch::getScopedObjectByGhostID(%ghostID, %type, %args);
			if(!isObject(%object)) {
				warn("Scratch Client " @ $Scratch::NetVersion @ " | WARNING IN BLOCK_DELETE: Ghost ID '" @ getWord(%line, 0) @ "' is not scoped yet");
				return;
			}
			Scratch::deleteBlock(%object, 1);
		case "BLOCK_UPDATE": if($CACHE::Scratch::Net_Server::Hosting) return;
			%ghostID   = getWord(%line, 0);
			%position  = getWords(%line, 1, 2);
			%extent    = getWords(%line, 3, 4);
			%args      = strReplace(getWord(%line, 5), "[[SPACE]]", " ");
			%type      = getWord(%line, 6);
			%createdBy = trim(stripMLControlChars(getWords(%line, 7, 30)));
			%object    = Scratch::getScopedObjectByGhostID(%ghostID, %type, %args);
			if(!isObject(%object)) {
				warn("Scratch Client " @ $Scratch::NetVersion @ " | WARNING IN BLOCK_UPDATE: Cannot scope object with Ghost ID '" @ getWord(%line, 0) @ "'");
				return;
			}
			%object.createdBy = %createdBy;
			%object.resize(getWord(%position, 0), getWord(%position, 1), getWord(%extent, 0), getWord(%extent, 1));
		case "BLOCK_ATTACH":
			%object = Scratch::getScopedObjectByGhostID(getWord(%line, 0));
			%other  = Scratch::getScopedObjectByGhostID(getWord(%line, 1));
			%side   = getWord(%line, 2);
			if(!isObject(%object)) {
				warn("Scratch Client " @ $Scratch::NetVersion @ " | WARNING IN BLOCK_ATTACH: Ghost ID '" @ getWord(%line, 0) @ "' is not scoped yet");
				return;
			}
			if(!isObject(%other)) {
				warn("Scratch Client " @ $Scratch::NetVersion @ " | WARNING IN BLOCK_ATTACH: Ghost ID '" @ getWord(%line, 1) @ "' is not scoped yet");
				return;
			}
			Scratch::attachBlock(%object, %other, %side, 1);
		case "BLOCK_DETACH":
			%object = Scratch::getScopedObjectByGhostID(getWord(%line, 0));
			%side   = getWord(%line, 1);
			if(!isObject(%object)) {
				warn("Scratch Client " @ $Scratch::NetVersion @ " | WARNING IN BLOCK_DETACH: Ghost ID '" @ getWord(%line, 0) @ "' is not scoped yet");
				return;
			}
			Scratch::detachBlock(%object, %side, 1);
		case "BLOCK_ATTACH_INPUTBLOCK":
			%ghostID      = getWord(%line, 0);
			%otherGhostID = getWord(%line, 1);
			%argIndex     = getWord(%line, 2);
			%object = Scratch::getScopedObjectByGhostID(%ghostID);
			%other  = Scratch::getScopedObjectByGhostID(%otherGhostID);
			if(!isObject(%object) || !isObject(%other)) return;
			if(!isObject(%blockArg = %other.blockArgObject[%argIndex])) return;
			if(%blockArg.argType $= "TEXTINPUT") return;
			Scratch::attachBlock(%object, %blockArg, "inside", 1);
		case "BLOCK_UPDATE_INPUT": if($CACHE::Scratch::Net_Server::Hosting) return;
			%ghostID = getWord(%line, 0);
			%object  = Scratch::getScopedObjectByGhostID(%ghostID);
			if(!isObject(%object)) {
				warn("Scratch Client " @ $Scratch::NetVersion @ " | WARNING IN BLOCK_UPDATE_INPUT: Ghost ID '" @ %ghostID @ "' is not scoped yet");
				return;
			}
			for(%i=0;%i<getWordCount(%line);%i++) {
				%blockArg = %object.blockArgObject[%i];
				if(!isObject(%blockArg)) break;
				%inputVal = getWord(%line, %i);
				if(getSubStr(%inputVal, 0, 1) $= "_") {
					%inputVal = getSubStr(%inputVal, 1, strLen(%inputVal));
					if(!isObject(%txtEd = %blockArg.textEdit)) continue;
					%txtEd.setValue(strReplace(%inputVal, "[SP]", " "));
					continue;
				}
				%child = Scratch::getScopedObjectByGhostID(%inputVal);
				if(!isObject(%child)) continue;
				Scratch::attachBlock(%child, %blockArg, "inside");
			}
		case "BLOCK_REMOVE_DELIMITER":
			%ghostID = getWord(%line, 0);
			%object  = Scratch::getScopedObjectByGhostID(%ghostID);
			if(!isObject(%object)) {
				warn("Scratch Client " @ $Scratch::NetVersion @ " | WARNING IN BLOCK_REMOVE_DELIMITER: Ghost ID '" @ %ghostID @ "' is not scoped yet");
				return;
			}
			Scratch::removeDelimiter(%object, 1);
		case "BLOCK_ADD_DELIMITER":
			%ghostID = getWord(%line, 0);
			%type    = getWord(%line, 1);
			%object  = Scratch::getScopedObjectByGhostID(%ghostID);
			if(!isObject(%object)) {
				warn("Scratch Client " @ $Scratch::NetVersion @ " | WARNING IN BLOCK_ADD_DELIMITER: Ghost ID '" @ %ghostID @ "' is not scoped yet");
				return;
			}
			Scratch::addDelimiter(%object, %type, 0, 1);
		case "BLOCK_UPDATE_ARGS":
			// <GhostID> <Type>[<Value>] <Type>[<Value>] <Type>[<Value>] etc
			%ghostID = getField(%line, 0);
			%object  = Scratch::getScopedObjectByGhostID(%ghostID);
			if(!isObject(%object)) {
				warn("Scratch Client " @ $Scratch::NetVersion @ " | WARNING IN BLOCK_UPDATE_ARGS: Ghost ID '" @ %ghostID @ "' is not scoped yet");
				return;
			}
			for(%i=1;%i<getFieldCount(%line);%i++) {
				%arg    = trim(stripMLControlChars(getField(%line, %i)));
				%data   = Scratch::splitStatus(%arg);
				%argObj = %object.blockArgObject[%i-1];
				if(!isObject(%argObj)) {
					if(isObject(%object.delimAddObj)) {
						%cmd = %object.delimAddObj.command;
						%cmd = strReplace(%cmd, ");", ", 0, 1);");
						eval(%cmd);
						%argObj = %object.blockArgObject[%i-1];
					}
					if(!isObject(%argObj)) {
						warn("Scratch Client " @ $Scratch::NetVersion @ " | WARNING IN BLOCK_UPDATE_ARGS: Server sent us more args than block with Ghost ID \'" @ %ghostID @ "\' has (sent=" @ getWordCount(%line) @ ", available=" @ %object.blockArgCount @ ")");
						return;
					}
				}
				%type  = getField(%data, 0);
				%value = getField(%data, 1);
				switch$(%type) {
					case "TEXT":
						if(%argObj.argType $= "BLOCKINPUT") {
							warn("Scratch Client " @ $Scratch::NetVersion @ " | WARNING IN BLOCK_UPDATE_ARGS: Argument " @ %i @ " is invalid (Argument type is 'BLOCKINPUT' and server sent 'TEXT')");
							continue;
						}
						if(!isObject(%txtEd = %argObj.textEdit)) {
							error("Scratch Client " @ $Scratch::NetVersion @ " | Text edit object for argument # " @ %i @ " of block " @ %object @ " is missing...?");
							continue;
						}
						if(%txtEd.getValue() $= %value) continue;
						%txtEd.setValue(%value);
						Scratch::setTextBoxWidth(%txtEd, %argObj, 1);
					case "ID":
						if(%argObj.argType $= "TEXTINPUT") {
							warn("Scratch Client " @ $Scratch::NetVersion @ " | WARNING IN BLOCK_UPDATE_ARGS: Argument " @ %i @ " is invalid (Argument type is 'TEXTINPUT' and server sent an 'INPUTBLOCK')");
							continue;
						}
						%other = Scratch::getScopedObjectByGhostID(%value);
						if(!isObject(%other)) {
							warn("Scratch Client " @ $Scratch::NetVersion @ " | WARNING IN BLOCK_UPDATE_ARGS: Ghost ID '" @ %ghostID @ "' is not scoped yet; cannot add input block");
							continue;
						}
						if(%argObj.innerConnection_Child != %other) {
							Scratch::attachBlock(%other, %argObj, "inside", 1);
						}
					case "NULL":
				}
			}
		case "UPDATE_MOUSE":
			%userName = getField(%line, 0);
			%position = getField(%line, 1);
			%dispText = getField(%line, 2);
			%object = Scratch::getScopedMouse(%userName);
			%object.textObj.setText("<font:Franklin Gothic Demi:14>" @ getSubStr(%dispText, 0, 255));
			%object.resize(getWord(%position, 0), getWord(%position, 1), 16, 32);
			%object.textObj.resize(getWord(%position, 0)+16, getWord(%position, 1)+16, 150, 16);
			ScratchGui.pushToBack(%object);
			ScratchGui.pushToBack(%object.textObj);
			if(Scratch_RightClickMenu.isVisible()) ScratchGui.pushToBack(Scratch_RightClickMenu);
		case "REMOVE_MOUSE":
			echo("Scratch Client " @ $Scratch::NetVersion @ " | CDROP: " @ %line);
			%object = Scratch::getScopedMouse(%line, 1);
			if(!isObject(%object)) return;
			$CACHE::Scratch::Net_Client::ScopedMouse[getStringCRC(%line)] = "";
			%object.textObj.delete();
			%object.delete();
		case "CHATMSG":
			Scratch::CollabMenu_chatMsg(getSubStr(%line, 0, 255));
		case "DISCONNECT": %this.disconnect();
		case "ADDUSER":
			%clientId = getField(%line, 0);
			%userName = getField(%line, 1);
			if(!Scratch_CollabMenu_PlayerList.isRowActive(%clientId)) {
				Scratch_CollabMenu_PlayerList.addRow(%clientID, %userName);
			}
		case "REMOVEUSER":
			%clientId = getField(%line, 0);
			if(Scratch_CollabMenu_PlayerList.isRowActive(%clientId)) {
				Scratch_CollabMenu_PlayerList.removeRowById(%clientId);
			}
		default: echo("Scratch Client " @ $Scratch::NetVersion @ " | Unhandled net action \"" @ %act @ "\"");
	}
}

function ScratchNet::sendMouseInfo() {
	if(!ScratchClientTCP.connected) return;
	for(%i=0;%i<ScratchMouseGroup.getCount();%i++) {
		%object = ScratchMouseGroup.getObject(%i);
		ScratchGui.pushToBack(%object);
		ScratchGui.pushToBack(%object.textObj);
	}
	ScratchGui.pushToBack(Scratch_CollabMenu);
	if(Scratch_RightClickMenu.isVisible()) ScratchGui.pushToBack(Scratch_RightClickMenu);
	%curX = getWord(Canvas.getCursorPos(), 0);
	%curY = getWord(Canvas.getCursorPos(), 1);
	if(%curX SPC %curY !$= $CACHE::Scratch::Net_Client::LastMousePos) {
		$CACHE::Scratch::Net_Client::LastMousePos = %curX SPC %curY;
		ScratchClientTCP.sendData("MOUSEPOS " @ %curX SPC %curY);
	}
	cancel($CACHE::Scratch::Net_Client::MouseSch);
	$CACHE::Scratch::Net_Client::MouseSch = schedule(32, 0, eval, "ScratchNet::sendMouseInfo();");
}

// Client CMDs

function scratchClientCmdsetAdminLevel(%bool) {
	Scratch::CollabMenu_showAdmin(mClamp(%bool, 0, 1));
}