package ScratchNet_ServerScoping {
	function ScratchServerClientTCP::disconnect(%this) {
		%this.sendData("DISCONNECT");
		Parent::disconnect(%this);
	}
};

//$CACHE::Scratch::Net_Server::Name
//$CACHE::Scratch::Net_Server::Password
//$CACHE::Scratch::Net_Server::MaxUsers

if(!isObject(ScratchServerTCP)) new TCPObject(ScratchServerTCP);
if(!isObject(ScratchServerClientGroup)) new SimGroup(ScratchServerClientGroup);

function Scratch::startListenServer(%port) {
	if(%port $= "") %port = "28000";
	ScratchServerTCP.listen(%port);
	echo("Scratch Server " @ $Scratch::NetVersion @ " | Listening for clients on port " @ %port @ (%pass ? " with password " @ $CACHE::Scratch::Net_Server::Password : ""));
	activatePackage(ScratchNet_ServerScoping);
	$CACHE::Scratch::Net_Server::Hosting = 1;
	Scratch::clientDisconnect();
	$CACHE::Scratch::Net_Server::Hosting = 1;
	$CACHE::Scratch::Net_Server::HostPassword = getRandom(15000, 900000);
	$CACHE::Scratch::Net_Client::JoinPassword = $CACHE::Scratch::Net_Server::Password;
	Scratch::connectToServer("127.0.0.1", %port);
}

function ScratchServerTCP::onConnectRequest(%this, %ip, %socket) {
	if(isObject(%this.connection[%ip])) {
		echo("Scratch Server " @ $Scratch::NetVersion @ " | Duplicate client with IP " @ %ip);
		%this.connection[%ip].disconnect();
		%this.connection[%ip].delete();
	}
	echo("Scratch Server " @ $Scratch::NetVersion @ " | CADD: " @ %ip);
	%newObj = new TCPObject("ScratchServerClientTCP", %socket);
	%newObj.ip = %ip;
	ScratchServerClientGroup.add(%newObj);
	%this.connection[%ip] = %newObj;
}

function ScratchServerClientTCP::sendData(%this, %line) {
	%this.send(%line @ "\r\n");
}

function ScratchServerClientTCP::onDisconnect(%this) {
	echo("Scratch Server " @ $Scratch::NetVersion @ " | CDROP: " @ %this.ip);
	if(%this.authed) {
		ScratchNet::announce("<font:Verdana Bold:12>* " @ %this.netName @ " has left the group.");
		for(%i=0;%i<ScratchServerClientGroup.getCount();%i++) {
			%object = ScratchServerClientGroup.getObject(%i);
			%object.sendData("REMOVE_MOUSE " @ %this.netName);
			%this.sendData("REMOVE_MOUSE " @ %object.netName);
			%this.sendData("REMOVEUSER" SPC %object);
			%object.sendData("REMOVEUSER" SPC %this);
		}
	}
	%this.schedule(1, delete);
}

// 1 if error, 0 if no error
function ScratchServerClientTCP::getManipulationError(%this) {
	if(!%this.authed || ($CACHE::Scratch::Net_Server::WatchOnlyModeEnabled && !%this.isHost)) {
		%this.sendData("BLOCK_ERROR Watch-only mode is currently enabled, which means you cannot manipulate blocks right now.");
		%this.scopeAll();
		return 1;
	}
	return 0;
}

function ScratchServerClientTCP::getUserName(%this) {
	return %this.netName;
}

function ScratchServerClientTCP::onLine(%this, %line) {
	%act  = firstWord(%line);
	%line = removeWord(%line, 0);
	if(!%this.authed) {
		switch$(%act) {
			case "INIT": if(%this.init) return;
				%this.init = 1;
				if($CACHE::Scratch::Net_Server::Password $= "") { %this.sendData("STATUS_NAME"); }
				else %this.sendData("STATUS_PASS");
			case "PASS": if(!%this.init) return;
				if(%line !$= $CACHE::Scratch::Net_Server::Password) {
					%this.sendData("STATUS_INVALID");
					%this.disconnect();
					%this.delete();
				} else { %this.sendData("STATUS_NAME"); }
			case "NAME": if(!%this.init) return;
				%netName = stripMLControlChars(trim(%line));
				%orgName = %netName;
				%amount  = 0;
				if(%netName $= "") { %this.sendData("STATUS_INVALID"); return; }
				for(%i=0;%i<ScratchServerClientGroup.getCount();%i++) {
					%object = ScratchServerClientGroup.getObject(%i);
					if(%object.netName $= %netName) {
						%netName = %orgName SPC "(" @ %amount++ @ ")";
						%i = 0;
					}
				}
				%this.netName = %netName;
				%this.authed = 1;
				%this.status = 0;
				%this.sendData("STATUS_OK" SPC %this.netName);
				for(%i=0;%i<ScratchServerClientGroup.getCount();%i++) {
					%obj = ScratchServerClientGroup.getObject(%i);
					%this.sendData("ADDUSER" SPC %obj TAB %obj.netName);
					%obj.sendData("ADDUSER" SPC %this TAB %this.netName);
				}
				echo("   +- Name: " @ %this.netname);
				ScratchNet::announce("<font:Verdana Bold:12>* " @ %this.netName @ " connected.");
			case "DISCONNECT": %this.schedule(1, delete); return;
		}
		return;
	}
	switch$(%act) {
		case "SERVERCMD":
			%funcName = "scratchServerCmd" @ getSafeVariableName(getField(%line, 0));
			for(%i=1;%i<getFieldCount(%line);%i++) %args = (%args $= "" ? "" : %args @ ", ") @ getSafeArgument(getField(%line, %i));
			if(!isFunction(%funcName)) return;
			eval(%funcName @ "(" @ %this.getId() @ (%args $= "" ? "" : ", " @ %args) @ ");");
		case "HOST_ATTEMPT":
			if($CACHE::Scratch::Net_Server::HostPassword $= "") return;
			if(%line $= $CACHE::Scratch::Net_Server::HostPassword) {
				$CACHE::Scratch::Net_Server::HostPassword = "";
				ScratchNet::announce("<color:007700>* " @ %this.netname @ " is now Host (Auto)");
				%this.isHost = 1;
				%this.sendData("CLIENTCMD setAdminLevel\t1");
			}
		case "CLEAR_CODE":
			if(!%this.isHost) {
				%this.sendData("BLOCK_ERROR You have insufficient permissions to clear all code.");
				return;
			}
			for(%i=0;%i<ScratchServerClientGroup.getCount();%i++) {
				ScratchServerClientGroup.getObject(%i).sendData("CLEAR_CODE");
			}
		case "REQUEST_GHOST_INFO":
			%this.scopeAll();
		case "CHATMSG":
			%line    = trim(stripMLControlChars(%line));
			if(%line $= "") return;
			%netname = %this.netName;
			%message = "<color:" @ (%this.isHost ? "007700" : "000000") @ "><font:Verdana Bold:12>" @ %netName @ ":<font:Verdana:12> " @ %line;
			echo("Scratch Server " @ $Scratch::NetVersion @ " | " @ stripMLControlChars(%message));
			for(%i=0;%i<ScratchServerClientGroup.getCount();%i++) {
				ScratchServerClientGroup.getObject(%i).sendData("CHATMSG " @ %message);
			}
		case "DISCONNECT": %this.schedule(1, delete); return;
		case "MOUSEPOS":
			%this.mousePos = %line;
			%this.scopeMousePos(%this.mousePos);
		case "BLOCK_DUPLICATE": if(%this.getManipulationError()) return;
			%object = Scratch::getObjectByGhostID(%line);
			if(!isObject(%object)) return;
			%newObj = Scratch::duplicateBlock(%object, 1);
			ScratchNet::scopeObject(%newObj);
		case "BLOCK_ATTACH_INPUTBLOCK": if(%this.getManipulationError()) return;
			%ghostID      = getWord(%line, 0);
			%otherGhostID = getWord(%line, 1);
			%argIndex     = getWord(%line, 2);
			%object = Scratch::getObjectByGhostID(%ghostID);
			%other  = Scratch::getObjectByGhostID(%otherGhostID);
			if(!isObject(%object) || !isObject(%other)) return;
			if(!isObject(%blockArg = %other.blockArgObject[%argIndex])) return;
			if(%blockArg.argType $= "TEXTINPUT") return;
			for(%i=0;%i<ScratchServerClientGroup.getCount();%i++) {
				ScratchServerClientGroup.getObject(%i).sendData("BLOCK_ATTACH_INPUTBLOCK" SPC %ghostID SPC %otherGhostID SPC %argIndex);
			}
		case "BLOCK_SPAWN": if(%this.getManipulationError()) return;
			%newObj = Scratch::createPremadeBlock(getField(%line, 0), getField(%line, 1));
			if(!isObject(%newObj)) return;
			echo("Scratch Server " @ $Scratch::NetVersion @ " | " @ %this @ "(" @ %this.netName @ ") spawned a block: " @ %newObj @ "<" @ %newObj.blockLayoutType @ ">");
			%newObj.createdBy = %this.netName;
			Scratch_Content.add(%newObj);
			Scratch::getGhostID(%newObj);
			ScratchNet::scopeObject(%newObj);
		case "BLOCK_DELETE": if(%this.getManipulationError()) return;
			%object = Scratch::getObjectByGhostID(%line);
			if(!isObject(%object)) return;
			echo("Scratch Server " @ $Scratch::NetVersion @ " | " @ %this @ "(" @ %this.netName @ ") deleted a block: " @ %object);
			for(%i=0;%i<ScratchServerClientGroup.getCount();%i++) {
				ScratchServerClientGroup.getObject(%i).sendData("BLOCK_DELETE " @ Scratch::getGhostID(%object));
			}
			Scratch::deleteBlock(%object, 1);
		case "BLOCK_MOVE": if(%this.getManipulationError()) return;
			%object = Scratch::getObjectByGhostID(getWord(%line, 0));
			if(!isObject(%object)) return;
			%xPos   = getWord(%line, 1);
			%yPos   = getWord(%line, 2);
			%object.resize(%xPos, %yPos, %object.getWidth(), %object.getHeight());
			ScratchNet::scopeObject(%object);
		case "BLOCK_ATTACH": if(%this.getManipulationError()) return;
			%object = Scratch::getObjectByGhostID(getWord(%line, 0));
			%other  = Scratch::getObjectByGhostID(getWord(%line, 1));
			if(!isObject(%object) || !isObject(%other)) return;
			for(%i=0;%i<ScratchServerClientGroup.getCount();%i++) {
				ScratchServerClientGroup.getObject(%i).sendData("BLOCK_ATTACH" SPC getWords(%line, 0, 2));
			}
		case "BLOCK_DETACH": if(%this.getManipulationError()) return;
			%object = Scratch::getObjectByGhostID(getWord(%line, 0));
			if(!isObject(%object)) return;
			for(%i=0;%i<ScratchServerClientGroup.getCount();%i++) {
				ScratchServerClientGroup.getObject(%i).sendData("BLOCK_DETACH" SPC getWords(%line, 0, 1));
			}
		case "BLOCK_SETTEXT": if(%this.getManipulationError()) return;
			%object = Scratch::getObjectByGhostID(getWord(%line, 0));
			if(!isObject(%object)) return;
			%argIdx = getWord(%line, 1);
			%text   = removeWord(removeWord(%line, 0), 0);
			%child  = %object.blockArgObject[%argIdx];
			if(!isObject(%child) || !isObject(%txtEd = %child.textEdit)) return;
			%txtEd.setValue(%text);
			ScratchNet::scopeObject(%object);
			Scratch::setTextBoxWidth(%txtEd, %child, 1);
		case "BLOCK_ADD_DELIMITER": if(%this.getManipulationError()) return;
			%object = Scratch::getObjectByGhostID(getWord(%line, 0));
			%type   = getWord(%line, 1);
			if(!isObject(%object)) return;
			for(%i=0;%i<ScratchServerClientGroup.getCount();%i++) {
				ScratchServerClientGroup.getObject(%i).sendData("BLOCK_ADD_DELIMITER" SPC getWord(%line, 0) SPC %type);
			}
		case "BLOCK_REMOVE_DELIMITER": if(%this.getManipulationError()) return;
			%object = Scratch::getObjectByGhostID(getWord(%line, 0));
			if(!isObject(%object)) return;
			for(%i=0;%i<ScratchServerClientGroup.getCount();%i++) {
				ScratchServerClientGroup.getObject(%i).sendData("BLOCK_REMOVE_DELIMITER" SPC getWord(%line, 0));
			}
		default: echo("Scratch Server " @ $Scratch::NetVersion @ " | Unhandled net action \"" @ %act @ "\"");
	}
}

function ScratchNet::announce(%msg) {
	for(%i=0;%i<ScratchServerClientGroup.getCount();%i++) {
		ScratchServerClientGroup.getObject(%i).sendData("CHATMSG " @ %msg);
	}
}

function Scratch::getObjectByGhostID(%ghostID) {
	if(!isObject(%this = $CACHE::Scratch::Net_Server::ObjectForGhostID[%ghostID])) return 0;
	return %this;
}

function Scratch::getGhostID(%this) {
	if(!isObject(%this)) return -1;
	if(%this.ghostID !$= "") return %this.ghostID;
	%this.ghostID = $CACHE::Scratch::Net_Server::HighestGhostID++;
	$CACHE::Scratch::Net_Server::GhostIDFor[%this] = %this.ghostID;
	$CACHE::Scratch::Net_Server::ObjectForGhostID[%this.ghostID] = %this;
	$CACHE::Scratch::Net_Client::ScopedObjForID[%this.ghostID] = %this;
}

function ScratchServerClientTCP::scopeMousePos(%client, %pos) {
	// Generate their mouse display text
	switch(%client.status) {
		case 0: %text = %client.netName; // Moving about
		case 1: %text = "<color:555555>" @ %client.netName; // Idle
	}
	if(%client.isHost) %text = "<color:FF00FF>" @ %client.netName @ "<bitmap:Add-Ons/Client_Scratch/img/crown>";
	for(%i=0;%i<ScratchServerClientGroup.getCount();%i++) {
		%object = ScratchServerClientGroup.getObject(%i);
		if(%object == %client) continue;
		%object.sendData("UPDATE_MOUSE " @ %client.netname TAB %pos TAB %text);
	}
}
function ScratchNet::scopeObject(%object) {
	if(!isObject(%object)) return;
	for(%i=0;%i<ScratchServerClientGroup.getCount();%i++) {
		ScratchServerClientGroup.getObject(%i).scopeObject(%object);
		ScratchServerClientGroup.getObject(%i).scopeObjectArgs(%object);
	}
}

function ScratchServerClientTCP::scopeAll(%this) {
	if(!%this.authed || !%this.init) return;
	%group = ScratchContainer;
	for(%i=0;%i<%group.getCount();%i++) {
		%object = %group.getObject(%i);
		%this.scopeObject(%object);
	}
	for(%i=0;%i<%group.getCount();%i++) {
		%object = %group.getObject(%i);
		%this.scopeObjectArgs(%object);
		if(isObject(%obj = %object.topConnection)) {
			%this.sendData("BLOCK_ATTACH" SPC Scratch::getGhostID(%object) SPC Scratch::getGhostID(%obj) SPC "bottom");
		}
	}
}

function ScratchServerClientTCP::scopeObject(%this, %object) {
	if(!isObject(%object)) return;
	if(!%this.authed || !%this.init) return;
	%ghID = Scratch::getGhostID(%object);
	%data =           %ghID;
	%data = %data SPC %object.getPosition();
	%data = %data SPC %object.getExtent();
	%data = %data SPC strReplace(%object.blockCreationArgs, " ", "[[SPACE]]");
	%data = %data SPC %object.blockLayoutType;
	if(%object.createdBy !$= "") %data = %data SPC %object.createdBy;
	%this.sendData("BLOCK_UPDATE" SPC %data);
}

function ScratchServerClientTCP::scopeObjectArgs(%this, %object) {
	if(!isObject(%object)) return;
	if(!%this.authed || !%this.init) return;
	%ghID = Scratch::getGhostID(%object);
	%data = %ghID;
	for(%i=0;%i<%object.blockArgCount;%i++) {
		%child = %object.blockArgObject[%i];
		if(isObject(%child_child = %child.innerConnection_Child)) {
			%data = %data TAB "ID[" @ Scratch::getGhostID(%child_child) @ "]";
			continue;
		}
		if(!isObject(%txtEd = %child.textEdit) || %child.argType $= "BLOCKINPUT") { %data = %data TAB "NULL[]"; continue; }
		%data = %data TAB "TEXT[" @ %txtEd.getValue() @ "]";
	}
	%data = trim(%data);
	%this.sendData("BLOCK_UPDATE_ARGS" SPC %data);
}