function ScratchPlayerGui::onWake(%this) {
	if(!$CACHE::Scratch::PlayerCreator::Init) {
		$CACHE::Scratch::PlayerCreator::Init = 1;
		Scratch_PlrCreator::setModel("base/data/shapes/player/m.dts");
		Scratch_PlrCreator_MaxHP.setValue("100");
		Scratch_PlrCreator_ModelFile.setValue("base/data/shapes/player/m.dts");
		
	}
}

function Scratch_PlrCreator::clearAll(%confirm) {
	if(!%confirm) {
		messageBoxYesNo("Confirmation - Scratch Editor", "Are you sure you want to clear the workspace?", "Scratch_PlrCreator::clearAll(1);");
		return;
	}
	Scratch_PlrCreator_AnimList.deleteAll();
	Scratch_PlrCreator_UIName.setValue("");
	Scratch_PlrCreator::setModel("base/data/shapes/player/m.dts");
	Scratch_PlrCreator_MaxHP.setValue("100");
	Scratch_PlrCreator_ModelFile.setValue("base/data/shapes/player/m.dts");
}

function Scratch_PlrCreator::setModel(%model) {
	if(!isFile(%model)) {
		error("ERROR: Scratch_PlrCreator::setModel() - File \"" @ %model @ "\" does not exist");
		return 0;
	}
	if(fileExt(%model) !$= ".dts") {
		error("ERROR: Scratch_PlrCreator::setModel() - Invalid file extension (got \"" @ fileExt(%model) @ "\", expected \".dts\")");
		return 0;
	}
	// mCameraRot = 0.030000 0.600000  2.430001
	// mOrbitPos  = 0.0      0.345002 -0.56496
	// mOrbitDist = 1.287905
	%this  = Scratch_PlrCreator_Preview;
	%animG = Scratch_PlrCreator_AnimList;
	%animG.deleteAll();
	
	// Clear the object
	%this.setEmpty();
	
	// Set the object
	%this.setObject("player", %model, "", "32");
	
	// Set the camera info
	%this.setCameraRot(0.030000, 0.600000, 2.430001);
	%this.setOrbitDist(1.287905);
	
	return 1;
}

function Scratch_PlrCreator::exportPlayertype(%fileName) {
	%modelFile    = Scratch_PlrCreator_ModelFile.getValue();
	%abilityGroup = Scratch_PlrCreator_AbilityList;
	if(fileExt(%fileName) !$= ".cs") {
		error("ERROR: Scratch_PlrCreator::exportPlayertype() - File extension must be '.cs'");
		return 0;
	}
	if(!isFile(%modelFile)) {
		error("ERROR: Scratch_PlrCreator::exportPlayertype() - Cannot find model file at \"" @ %fileName @ "\"");
		return 0;
	}
	// Set the variables
	%objectName = getSafeVariableName(fileBase(%modelFile));

	// Generate the whitespace (to make it look nice)
	for(%i=0;%i<strLen(Scratch_PlrCreator_AnimList.getCount());%i++) %ws = %ws @ " ";
	
	// Open the file
	%SO = new FileObject();
	%SO.openForWrite(%fileName);
	
	// Write the TSShapeConstructor first
	%SO.writeLine("// Load dts shapes and merge animations");
	%SO.writeLine("datablock TSShapeConstructor(" @ %objectName @ "Dts)");
	%SO.writeLine("{");
	%SO.writeLine("\tbaseShape = \"" @ %modelFile @ "\";");
	%SO.writeLine("\t");
	for(%i=0;%i<Scratch_PlrCreator_AnimList.getCount();%i++) {
		%object = Scratch_PlrCreator_AnimList.getObject(%i);
		%SO.writeLine("\tsequence" @ %i @ getSubStr(%ws, strLen(%i)-1, strLen(%ws)) @ "= \"" @ %object.dsqFile SPC getWord(%object.getObject(2).getValue(), 0) @ "\";");
	}
	%SO.writeLine("};");

	// Generate the whitespace (to make it look nice)
	%ws      = "";
	%biggest = strLen("shapeFile");
	for(%i=0;%i<%abilityGroup.getCount();%i++) {
		%ability = %abilityGroup.getObject(%i);
		%varName = %ability.varName;
		if(strLen(%varName) >= %biggest) %biggest = strLen(%varName);
	}
	for(%i=0;%i<%biggest;%i++) %ws = %ws @ " ";

	// Next, write the object declaration
	%SO.writeLine("datablock PlayerData(" @ %objectName @ "Armor : PlayerStandardArmor)");
	%SO.writeLine("{");
	%SO.writeLine("\tshapeFile" @ getSubStr(%ws, strLen("shapeFile")-1, strLen(%ws)) @ "= \"" @ Scratch_PlrCreator_ModelFile.getValue() @ "\";");
	%SO.writeLine("\tuiName" @ getSubStr(%ws, strLen("uiName")-1, strLen(%ws)) @ "= \"" @ Scratch_PlrCreator_UIName.getValue() @ "\";");
	%SO.writeLine("\tmaxDamage" @ getSubStr(%ws, strLen("maxDamage")-1, strLen(%ws)) @ "= \"" @ Scratch_PlrCreator_MaxHP.getValue() @ "\";");
	// Write abilities
	for(%i=0;%i<%abilityGroup.getCount();%i++) {
		%ability = %abilityGroup.getObject(%i);
		%varName = %ability.varName;
		%value   = %ability.varValue;
		%SO.writeLine("\t" @ %varName @ getSubStr(%ws, strLen(%varName)-1, strLen(%ws)) @ "= " @ getSafeArgument(%value) @ ";");
	}
	
	// End the object definition
	%SO.writeLine("};");
	
	// We are done with the file object
	%SO.close();
	%SO.delete();
	
	// Return success
	return 1;
}

function Scratch_PlrCreator::loadAnimation(%file) {
	if(!isFile(%file)) {
		error("ERROR: Scratch_PlrCreator::loadAnimation() - File \"" @ %file @ "\" does not exist");
		return 0;
	}
	if(fileExt(%file) !$= ".dsq") {
		error("ERROR: Scratch_PlrCreator::loadAnimation() - Invalid file extension (got \"" @ fileExt(%file) @ "\", expected \".dsq\")");
		return 0;
	}
	Scratch_PlrCreator_Preview.loadDSQ("player", %file);
	Scratch_PlrCreator::registerAnimation(%file);
	%newObj = Scratch_PlrCreator_AnimList.getObject(Scratch_PlrCreator_AnimList.getCount()-1);
	%newObj.getObject(2).setValue(%animName);
}

function Scratch_PlrCreator::importScript(%file) {
	if(!isFile(%file)) {
		error("ERROR: Scratch_PlrCreator::importScript() - File \"" @ %file @ "\" does not exist");
		return 0;
	}
	if(fileExt(%file) !$= ".cs") {
		error("ERROR: Scratch_PlrCreator::importScript() - Invalid file extension (got \"" @ fileExt(%file) @ "\", expected \".cs\")");
		return 0;
	}
	// Set variables
	%group = Scratch_PlrCreator_AbilityList;
	%animG = Scratch_PlrCreator_AnimList;
	%animG.deleteAll();
	
	// Create a file object
	%SO = new FileObject();
	%SO.openForRead(%file);
	
	// Check the file for validity first
	while(!%SO.isEOF()) {
		%line = %SO.readLine();
		if(strIPos(%line, "datablock TSShapeConstructor(") != -1) %foundTSShapeConst = 1;
		if(strIPos(%line, "datablock PlayerData(") != -1) %foundPlayerData = 1;
	}
	%SO.close();
	
	if(!%foundPlayerData) {
		%SO.delete();
		error("ERROR: Scratch_PlrCreator::importScript() - Couldn't find PlayerData in file \"" @ %file @ "\"");
		messageBoxOK("ERROR - Scratch Editor", "This is an invalid playertype definition file (Unable to find PlayerData).");
		return 0;
	}
	if(!%foundTSShapeConst) {
		%SO.delete();
		warn("WARNING: Scratch_PlrCreator::importScript() - Couldn't find TSShapeConstructor in file \"" @ %file @ "\"");
		messageBoxOK("ERROR - Scratch Editor", "Animations for this player type could not be loaded (Unable to find TSShapeConstructor).");
		return 0;
	}
	
	// Now parse the file
	%SO.openForRead(%file);
	while(!%SO.isEOF()) {
		%line = trim(%SO.readLine());
		if(%line $= "") continue;
		if(%inTSConstructor) {
			if(%line $= "{") continue;
			if(%line $= "};") { %inTSConstructor = 0; continue; }
			%varName  = getWord(%line, 0);
			%varValue = trim(removeWord(trim(removeWord(%line, 0)), 0));
			if(getSubStr(%varValue, 0, 1) $= "\"" && getSubStr(%varValue, strLen(%varValue)-2, 2) $= "\";") %varValue = getSubStr(%varValue, 1, strLen(%varValue)-3);
			%animName = getWord(%varValue, getWordCount(%varValue)-1);
			%animFile = removeWord(%varValue, getWordCount(%varValue)-1);
			%animFile = strReplace(%animFile, "./", filePath(%file) @ "/");
			%anim[-1+%animCount++] = %animFile TAB %animName;
			continue;
		}
		if(%inPlayerData) {
			if(%line $= "{") continue;
			if(%line $= "};") break;
			switch$(getWord(%line, 0)) {
				case "shapeFile":
					%string = trim(removeWord(trim(removeWord(%line, 0)), 0));
					if(getSubStr(%string, 0, 1) $= "\"" && getSubStr(%string, strLen(%string)-2, 2) $= "\";") %string = getSubStr(%string, 1, strLen(%string)-3);
					else %string = getSubStr(%string, 0, strLen(%string)-1);
					%string = strReplace(%string, "./", filePath(%file) @ "/");
					Scratch_PlrCreator::setModel(%string);
					Scratch_PlrCreator_ModelFile.setValue(%string);
				case "uiName":
					%string = trim(removeWord(trim(removeWord(%line, 0)), 0));
					if(getSubStr(%string, 0, 1) $= "\"" && getSubStr(%string, strLen(%string)-2, 2) $= "\";") %string = getSubStr(%string, 1, strLen(%string)-3);
					else %string = getSubStr(%string, 0, strLen(%string)-1);
					Scratch_PlrCreator_UIName.setValue(%string);
				case "maxDamage":
					%string = trim(removeWord(trim(removeWord(%line, 0)), 0));
					if(getSubStr(%string, 0, 1) $= "\"" && getSubStr(%string, strLen(%string)-2, 2) $= "\";") %string = getSubStr(%string, 1, strLen(%string)-3);
					else %string = getSubStr(%string, 0, strLen(%string)-1);
					Scratch_PlrCreator_MaxHP.setValue(%string);
				default:
					%string = trim(removeWord(trim(removeWord(%line, 0)), 0));
					if(getSubStr(%string, 0, 1) $= "\"" && getSubStr(%string, strLen(%string)-2, 2) $= "\";") %string = getSubStr(%string, 1, strLen(%string)-3);
					else %string = getSubStr(%string, 0, strLen(%string)-1);
					for(%i=0;%i<%group.getCount();%i++) {
						%object = %group.getObject(%i);
						if(getWord(%line, 0) !$= %object.varName) continue;
						switch$(%object.varType) {
							case "bool":  %object.setValue(mClamp(%string, 0, 1));
							case "text":  %object.setValue(%string);
							case "float": %object.setValue(atof(%string));
						}
						break;
					}
			}
			continue;
		}
		if(strIPos(%line, "datablock TSShapeConstructor(") != -1) %inTSConstructor = 1;
		if(strIPos(%line, "datablock PlayerData(") != -1) %inPlayerData = 1;
	}
	%SO.close();
	
	// Delete and return success
	%SO.delete();

	// Load animations
	for(%i=0;%i<%animCount;%i++) {
		%animFile = getField(%anim[%i], 0);
		%animName = getField(%anim[%i], 1);
		if(%animFile $= "") continue;
		Scratch_PlrCreator::loadAnimation(%animFile);
		%newObj = Scratch_PlrCreator_AnimList.getObject(Scratch_PlrCreator_AnimList.getCount()-1);
		%newObj.getObject(2).setValue(%animName);
	}
	return 1;
}

function ScratchPlrSel_ModelFile(%path, %fileName) {
	switch$(fileExt(%fileName)) {
		case ".dts":
			Scratch_PlrCreator_ModelFile.setValue(%path @ "/" @ %fileName);
			Scratch_PlrCreator::setModel(%path @ "/" @ %fileName);
		case ".cs":
			Scratch_PlrCreator::importScript(%path @ "/" @ %fileName);
	}
}



function Scratch_PlrCreator::setModel_DoFileSel(%this) {
	ScratchGui.openDirectorySelector("ScratchPlrSel_ModelFile", "Select", "_.dts_ Torque Model (.dts)\t_.cs_ Torque Script (.cs)");
}

function ScratchPlrSel_MountFile(%path, %fileName) {
	Scratch_PlrCreator_MountFile.setValue(%path @ "/" @ %fileName);
	Scratch_PlrCreator::setMountModel(%path @ "/" @ %fileName);
}

function Scratch_PlrCreator::setMount_DoFileSel(%this) {
	ScratchGui.openDirectorySelector("ScratchPlrSel_MountFile", "Select", "_.dts_ Torque Model (.dts)");
}

function ScratchPlrSel_addAnimCallback(%path, %fileName) {
	Scratch_PlrCreator::loadAnimation(%path @ "/" @ %fileName);
}

function Scratch_PlrCreator::addAnimation() {
	ScratchGui.openDirectorySelector("ScratchPlrSel_addAnimCallback", "Select", "_.dsq_ Torque Animation (.dsq)");
}

function ScratchPlrSel_exportCallback(%path, %fileName) {
	Scratch_PlrCreator::exportPlayertype(%path @ "/" @ %fileName);
}

function Scratch_PlrCreator::exportAnimBtn() {
	ScratchGui.openDirectorySelector("ScratchPlrSel_exportCallback", "Export", "_.cs_ Torque Script (.cs)");
}

function Scratch_PlrCreator::removeAnimation() {
	if(!isObject(%obj = $CACHE::Scratch::PlayerCreator::SelectedObject)) return;
	%obj.delete();
	for(%i=0;%i<Scratch_PlrCreator_AnimList.getCount();%i++) {
		%object          = Scratch_PlrCreator_AnimList.getObject(%i);
		%defColor        = (isEven(%i) ? "0 100 176 50" : "255 255 255 255");
		%color           = %defColor;
		%object.color    = %color;
		%object.defColor = %defColor;
		%object.resize(0, 28*%i, Scratch_PlrCreator_AnimList.getWidth(), 28);
	}
	Scratch_PlrCreator_AnimList.resize(1, 1, Scratch_PlrCreator_AnimList.getWidth(), 28*Scratch_PlrCreator_AnimList.getCount());
}

