package ScratchRenderPackage {
	function GuiControl::resize(%this, %x, %y, %w, %h) {
		Parent::resize(%this, %x, %y, %w, %h);
		if(ScratchContainer.isMember(%this) && isEventPending($CACHE::Scratch::RenderSystem::TreeRenderSchedule)) {
			Scratch::update(%this);
			%parent = Scratch::getHighestTopConnection(Scratch::getHighestInnerConnection(%this));
			Scratch::update(%parent);
		}
	}
};
activatePackage(ScratchRenderPackage);

function treeEcho(%msg) {
	$CACHE::TreeEcho::Amt = mClamp($CACHE::TreeEcho::Amt,0,50);
	%amt = $CACHE::TreeEcho::Amt;
	for(%i=0;%i<%amt;%i++) %pre = %pre @ "  |";
	if(getSubStr(%msg,0,1) $= "+") { $CACHE::TreeEcho::Amt++; if(%pre !$= "") %pre = getSubStr(%pre,0,strLen(%pre)-1); }
	if(getSubStr(%msg,0,1) $= "-") { $CACHE::TreeEcho::Amt--; return; }
	echo(%pre @ %msg);
}

function Scratch::renderDebugRect(%name, %X, %Y, %W, %H) {
	if(!$Scratch::VerboseDebug) return;
	%name = getSafeVariableName(%name);
	if(!isObject(%obj = ("Scratch_Debug" @ %name))) {
		%obj = new GuiSwatchCtrl("Scratch_Debug" @ %name) {
			profile = "GuiScratchNoModalProfile";
			color   = "255 0 0 40";
		};
		Scratch_Content.add(%obj);
	}
	%obj.setVisible(1);
	%obj.resize(%X, %Y, %W, %H);
	cancel(%obj.timeoutDissapear);
	%obj.timeoutDissapear = %obj.schedule(2000, setVisible, 0);
}

function Scratch::debugEcho(%msg) {
	if(!$Scratch::VerboseDebug) return;
	$CACHE::TreeEcho::Amt = mClamp($CACHE::TreeEcho::Amt,0,50);
	%amt = $CACHE::TreeEcho::Amt;
	for(%i=0;%i<%amt;%i++) %pre = %pre @ "  |";
	if(getSubStr(%msg,0,1) $= "+") { $CACHE::TreeEcho::Amt++; if(%pre !$= "") %pre = getSubStr(%pre,0,strLen(%pre)-1); }
	if(getSubStr(%msg,0,1) $= "-") { $CACHE::TreeEcho::Amt--; return; }
	echo("\c4[SCRATCH DEBUG]\c6 :: " @ %pre @ %msg);
}

function ScratchContainerBlock::onRemove(%this) {
	if(isObject(%this.debugTextObject)) %this.debugTextObject.delete();
}

function Scratch::splitStatus(%status) {
	if(strIPos(%status, "[") != -1) {
		%stArgs = getSubStr(%status, strIPos(%status, "[")+1, strLen(%status));
		%stArgs = getSubStr(%stArgs, 0, strLen(%stArgs)-1);
		%status = getSubStr(%status, 0, strIPos(%status, "["));
	}
	return %status TAB %stArgs;
}

function Scratch::update(%this, %amt) {
	if(Scratch::getBlockType(%this) !$= "BLOCK") {
		error("ERROR: Scratch::update()  Failed to update non-block object");
		return 0;
	}
	if(isObject(%p = %this.innerConnection_Parent)) Scratch::update(%p.getGroup(), %amt);
	if($Scratch::RenderType && !isWordInString($CACHE::Scratch::Renderer::UpdateList, %this.getID())) $CACHE::Scratch::Renderer::UpdateList = trim($CACHE::Scratch::Renderer::UpdateList SPC %this.getID());
	%status = Scratch::splitStatus(%this.status);
	%stArgs = getField(%status, 1);
	%status = getField(%status, 0);
	switch$(%status) {
		case "FOLLOW":
		case "READY": %this.status = "UPDATE" @ (%amt !$= "" ? "[" @ atoi(%amt) @ "]" : "");
		case "DRAG":
		case "UPDATE":
		default: error("ERROR: Scratch::update() - Unimplemented status '" @ %status @ "'.");
	}
}

function Scratch::renderBlockDebug(%this) {
	if(!$Scratch::Debug) return;
	if(!isObject(%this)) return;
	if(!isObject(%dbgTxt = %this.debugTextObject)) {
		%this.debugTextObject = new GuiMLTextCtrl() {
			extent = %this.getWidth() SPC "12";
		};
		%dbgTxt = %this.debugTextObject;
		Scratch_Content.add(%dbgTxt);
	}
	%txt =        "<font:Consolas:12><color:777777>ID_" @ %this.getID() @ "<br>";
	%txt = %txt @ "(" @ %this.getLeft() @ "," @ %this.getTop() @ ")<br>";
	%txt = %txt @ "<" @ %this.getWidth() @ "," @ %this.getHeight() @ "><br>";
	%txt = %txt @ (isObject(%this.topConnection) ? "<color:ff0000>" : "<color:00ff00>") @ "TOP<br>";
	%txt = %txt @ (isObject(%this.bottomConnection) ? "<color:ff0000>" : "<color:00ff00>") @ "BTM<br>";
	%txt = %txt @ (isObject(%this.innerConnection_Parent) ? "<color:ff0000>" : "<color:00ff00>") @ "I_P<br>";
	%txt = %txt @ (isObject(%this.innerConnection_Child) ? "<color:ff0000>" : "<color:00ff00>") @ "I_C<color:000000><br>";
	if(getField(Scratch::splitStatus(%this.status), 0) $= "FOLLOW")
		%txt = %txt @ %this.status @ "=" @ Scratch::handleFollowStatus(getField(Scratch::splitStatus(%this.status), 1).getGroup());
	else %txt = %txt @ %this.status;
	%dbgTxt.setText(%txt);
	%this.getGroup().getGroup().bringToFront(%dbgTxt);
	if(isObject(%b = %this.bottomConnection)) {
		if(isObject(%t = %this.topConnection))
			%dbgTxt.resize(%this.getLeft()-80, %this.getTop(), 80, %dbgTxt.getHeight());
		else %dbgTxt.resize(%this.getRight(), %this.getTop(), %this.getWidth(), %dbgTxt.getHeight());
	} else %dbgTxt.resize(%this.getLeft(), %this.getBottom(), %this.getWidth(), %dbgTxt.getHeight());
}

function Scratch::renderDebugOverlay() {
	if(!$Scratch::Debug) return;
	if(!isObject(%dbgTxt = Scratch_Content.debugTextObject)) {
		Scratch_Content.debugTextObject = new GuiMLTextCtrl() {
			extent = %this.getWidth() SPC "12";
		};
		%dbgTxt = Scratch_Content.debugTextObject;
		Scratch_Content.add(%dbgTxt);
	}
	%txt =        "<font:Consolas:12><color:777777>";
	%txt = %txt @ "ScratchContainer Count: " @ ScratchContainer.getCount() @ "<br>";
	%txt = %txt @ "Scratch_Content  Count: " @ Scratch_Content.getCount() @ "<br>";
	%txt = %txt @ "            RenderType: " @ $Scratch::RenderType @ "<br>";
	%txt = %txt @ "        Render Details: " @ ($Scratch::RenderType ? getWordCount($CACHE::Scratch::Renderer::UpdateList) : "n/a");
	%dbgTxt.setText(%txt);
	%dbgTxt.resize(0, 0, Scratch_Content.getWidth(), Scratch_Content.getHeight());
	Scratch_Content.bringToFront(%dbgTxt);
}

function Scratch::handleDragStatus(%this) {
	if(!isObject(%this)) {
		error("ERROR: Scratch::handleDragStatus() - Object " @ %this @ " does not exist.");
		return;
	}
	%newPos = getWord(Canvas.getCursorPos(), 0)-getWord(%this.mouseOffset, 0) SPC getWord(Canvas.getCursorPos(), 1)-getWord(%this.mouseOffset, 1);
	%this.resize(getWord(%newPos, 0), getWord(%newPos, 1), %this.getWidth(), %this.getHeight());
	%this.control.lastConnectPoint = %this.control.findNearestConnectionPoint();
	if(!$Pref::Scratch::DisplayPrefs::EnableOutlines) return;
	%LCP = %this.control.lastConnectPoint;
	if(trim(%LCP) !$= "") {
		%lyr = getField(%LCP, 0);
		%obj = getField(%LCP, 1);
		if(!isObject(%this.lastDragObj) || %this.lastDragObj.getId() != %obj.getId()) {
			%this.lastDragObj = %obj;
			ScratchHighlightContainer.deleteAll();
		}
		if(%lyr $= "back") {
			// Above
			Scratch::highlightBlock(%obj, 1);
		} else if(%lyr $= "front") {
			// Below
			Scratch::highlightBlock(%obj, 2);
		} else if(%lyr $= "inside") {
			// Inside
		} else if(%lyr $= "inside_top") {
			// Inside top
		}
		Scratch_Content.pushToBack(%this);
	} else {
		if(ScratchHighlightContainer.getCount() != 0) ScratchHighlightContainer.deleteAll();
	}
}

function Scratch::handleFollowStatus(%this, %getObject) {
	if(!isObject(%this)) return 0;
	if(Scratch::getBlockType(%this) !$= "BLOCK") {
		error("ERROR: Scratch::handleFollowStatus() - Invalid block type!");
		return 0;
	}
	%object = %this;
	//echo("\c4START");
	while(isObject(%object)) {
		if(%dbg++ >= 1000) { echo("Scratch::handleFollowStatus() [test.cs] -- Broke possible infinite loop"); break; }
		%status = Scratch::splitStatus(%object.status);
		%stArgs = getField(%status, 1);
		%status = getField(%status, 0);
		//echo((-1+%c++) SPC ":" SPC %object @ "(" @ Scratch::getBlockType(%object) @ ")");
		switch$(%status) {
			case "READY":  return (%getObject ? %object : 0);
			case "FOLLOW": if(!isObject(%stArgs)) { %this.status = "READY"; return 0; } %object = %stArgs.getGroup(); continue;
			default: return (%getObject ? %object : 1);
		}
	}
	return 0;
}

function Scratch::renderTree() {
	if($Scratch::RenderBoost) {
		cancel($CACHE::Scratch::RenderSystem::TreeRenderSchedule);
		Scratch_renderTree();
		return;
	}
	Scratch::renderTooltip();
	Scratch::renderRightClickMenu();
	Scratch::renderDebugOverlay();
	%this       = ScratchContainer;
	%isDragging = 0;
	// Sort the render list first
	if($CACHE::Scratch::Renderer::UpdateList $= "") {
		cancel($CACHE::Scratch::RenderSystem::TreeRenderSchedule);
		$CACHE::Scratch::RenderSystem::TreeRenderSchedule = schedule(1, 0, eval, "Scratch::renderTree();");
		return;
	}
	%renderList = $CACHE::Scratch::Renderer::UpdateList;
	%stage      = 0;
	%newGroupS  = Scratch_Content.getGroup().getWidth() SPC Scratch_Content.getGroup().getHeight();
	for(%i=0;%i<getWordCount(%renderList);%i++) {
		%obj       = getWord(%renderList, %i);
		if(%got[%obj]) continue;
		switch(%stage) {
			case 0:
				if(%obj.blockType !$= "INPUTBLOCK") {
					%newList    = trim(%newList SPC %obj);
					%renderList = removeWord(%renderList, %i);
					%got[%obj]  = 1;
					%i          = -1;
					continue;
				}
				if(%i == getWordCount(%renderList)-1) { %stage++; %i = -1; continue; }
			case 1: // Find all input blocks with a non-empty innerConnectionTable
				%groups = %obj;
				%debug  = 0;
				while(%groups !$= "") {
					if(%debug++ >= 1000) { error("ERROR: Scratch::renderTree() - Broke while loop"); return; }
					%object = getWord(%groups, 0);
					if(!isObject(%object)) break;
					if(%got[%object]) {
						%groups = removeWord(%groups, 0);
						continue;
					}
					%newList      = trim(%newList SPC %object);
					%got[%object] = 1;
					for(%i=0;%i<getWordCount(%object.innerConnectionTable);%i++) {
						%groups = trim(%groups SPC getWord(%object.innerConnectionTable, %i));
					}
					%groups = removeWord(%groups, 0);
				}
		}
	}
	$CACHE::Scratch::Renderer::UpdateList = %newList;
	$CACHE::Scratch::Renderer::Debug::LastUpdateList = %newList;
	// Now render the render list
	for(%i=0;%i<getWordCount($CACHE::Scratch::Renderer::UpdateList);%i++) {
		%object    = getWord($CACHE::Scratch::Renderer::UpdateList, %i);
		%status    = Scratch::splitStatus(%object.status);
		%stArgs    = getField(%status, 1);
		%status    = getField(%status, 0);
		%remove    = 1;
		%updateAmt = 1;
		switch$(%status) {
			case "READY":  continue;
			case "FOLLOW": if(!Scratch::handleFollowStatus(%object)) continue;
			case "UPDATE":
				if(%stArgs $= "") %stArgs = 1;
				%updateAmt = atoi(%stArgs);
				if(%stArgs == 1) %object.status = "READY";
				else Scratch::update(%object, %stArgs--);
			case "DRAG":   Scratch::handleDragStatus(%object); %remove = 0; %isDragging = 1;
		}
		Scratch::renderBlockDebug(%object);
		for(%j=0;%j<%updateAmt;%j++) {
			if(isFunction(ScratchContainerBlock, "renderBlock") && $Scratch::RenderBoost) {
				if(%object.container) Scratch::renderContainer(%object);
				else %object.renderBlock();
				continue;
			}
			Scratch::renderBlock(%object);
		}
		// Calculate the new scratch_content size
		if(!isObject(%object)) continue;
		if(%object.getRight() >= getWord(%newGroupS, 0)) %newGroupS = setWord(%newGroupS, 0, %object.getRight());
		if(%object.getBottom() >= getWord(%newGroupS, 1)) %newGroupS = setWord(%newGroupS, 1, %object.getBottom());
		if(%remove) { $CACHE::Scratch::Renderer::UpdateList = removeWord($CACHE::Scratch::Renderer::UpdateList, %i); %i--; }
	}
	if(Scratch_Content.getExtent() !$= %newGroupS) {
		Scratch_Content.resize(1, 1, getWord(%newGroupS, 0), getWord(%newGroupS, 1));
	}
	if(Scratch_Highlight.profile !$= "GuiScratchNoModalProfile") Scratch_Highlight.setProfile("GuiScratchNoModalProfile");
	%hlit = (%isDragging ? 1 : 0);
	if(!%hlit) {
		if(Scratch_Highlight.isVisible()) Scratch_Highlight.setVisible(0);
		//%gui = Scratch_Content.getGroup().getGroup();
		//if(!%gui.isMember(Scratch_Highlight)) %gui.add(Scratch_Highlight);
	} else {
		//%gui = Scratch_Content;
		//if(!%gui.isMember(Scratch_Highlight)) %gui.add(Scratch_Highlight);
	}
	cancel($CACHE::Scratch::RenderSystem::TreeRenderSchedule);
	$CACHE::Scratch::RenderSystem::TreeRenderSchedule = schedule(1, 0, eval, "Scratch::renderTree();");
}

function Scratch::renderBlock(%this) {
	if(!isObject(%this)) return;
	if($Scratch::renderBoost) { %this.renderBlock(); return; }
	if(%this.container) { Scratch::renderContainer(%this); return; }
	// Variables
	%ws      = 4; // Whitespace between blocks.
	%x       = (%this.top ? 26 : (%this.blockType $= "INPUTBLOCK" ? 3 : 9));
	%biggest = 0;

	// First, go to our top connection if we have one.
	if(isObject(%obj = %this.topConnection)) {
		//%this.resize(%obj.getLeft()+mAbs(%obj.bottomConnector.getLeft()-%this.topConnector.getLeft()), %obj.getBottom()-7, %this.getWidth(), %this.getHeight());
		Scratch_Content.bringToFront(%this);
	}
	Scratch::renderConnections(%this);

	// Or if we have a bottom connection, make that object come to us.
	if(isObject(%obj = %this.bottomConnection)) {
		%bc   = %this.bottomConnector;
		%tc   = %obj.topConnector;
		%newX = %this.getLeft()+(%bc.getLeft()-%tc.getLeft());
		if(%obj.getPosition() !$= %newX SPC %this.getBottom()-7) {
			%obj.resize(%newX, %this.getBottom()-7, %obj.getWidth(), %obj.getHeight());
			Scratch::update(%obj);
		}
	}
	// Push the control object first so that we can push arg boxes infront
	%this.pushToBack(%this.control);

	// Render our contents
	%ourH     = %this.getHeight()-(%this.bottom ? 7 : 0);
	%newH     = getWord(%this.minExtent, 1);
	for(%i=0;%i<%this.totalArgCount;%i++) {
		%child = %this.allArgObjects[%i];
		if(%child.ignore) continue; // .ignore is used mainly for background elements
		%lastChild = %child;
		%rectX     = (%child.static ? %child.getLeft() : %x); // If this object is static, then we'll only vertically align it.
		%rectY     = (%ourH/2)-(%child.getHeight()/2);
		if(%child.getPosition() $= %rectX SPC %rectY) continue; // This block is already correctly positioned.
		%child.resize(%rectX, %rectY, %child.getWidth(), %child.getHeight());
		if(%child.getHeight()+(%ws*2) > %newH) %newH = %child.getHeight()+(%this.bottom ? (%ws*2) : 0);
		%this.pushToBack(%child);
		%x += %child.getWidth()+%ws;
	}

	// Re-adjust the block size variables if neccessary
	if(isObject(%lastChild) && %lastChild.getRight()+%ws != %this.getWidth()) {
		%x               = %lastChild.getRight()+%ws;
		%forceUpdateSize = 1;
	}

	// Resize ourselves
	if((%x > %this.getWidth() || %newH != %ourH) || %forceUpdateSize) {
		//%newH  += (%this.bottom ? 7 : 0);
		%this.resize(%this.getLeft(), %this.getTop(), %x, %newH+(%ws*2));
	}

	// Render the arguments by argument index
	%count = -1;
	for(%i=0;%i<%this.getCount();%i++) {
		%object = %this.getObject(%i);
		if(%object.argIndex $= "") continue;
		%object[%object.argIndex] = %object;
		%count++;
	}
	for(%i=0;%i<%count+1;%i++) {
		%this.pushToBack(%object[%i]);
	}

	// Render our connections
	Scratch::renderConnections(%this);

	// Set our offsets
	Scratch::setBlockOffsets(%this);
}

function Scratch::renderConnections(%this, %resetHeight, %doNotResize) {
	if(!isObject(%this)) return;
	if(%this.container) { Scratch::renderContainerConnections(%this, %resetHeight, %doNotResize); return; }
	// Render the inner connections first
	%ws       = 4;
	%biggestY = (%resetHeight ? getWord(%this.minExtent, 1) : %this.getHeight());
	if(!%this.renderedBefore) {
		%this.renderedBefore = 1;
		for(%i=0;%i<%this.getCount();%i++) {
			%child = %this.getObject(%i);
			if(%child.ignore) continue;
			if(Scratch::getBlockType(%child) !$= "ARG_PLACEHOLDER") continue;
			Scratch::setBlockOffsets(%child);
		}
	}
	for(%i=0;%i<getWordCount(%this.innerConnectionTable);%i++) {
		%object = getWord(%this.innerConnectionTable, %i);
		%child  = %object.innerConnection_Parent;
		Scratch::setBlockOffsets(%child);
		%object.resize(%this.getLeft()+%child.getLeft(), %this.getTop()+%child.getTop(), %object.getWidth(), %object.getHeight());
		%child.resize(%child.getLeft(), %child.getTop(), %object.getWidth(), %object.getHeight());
		if(%child.getHeight()+(%ws*2) > %biggestY) %biggestY = %child.getHeight()+(%ws*2);
		%object.getGroup().pushToBack(%object);
		Scratch::renderBlock(%object);
	}
	if(!%doNotResize) %this.resize(%this.getLeft(), %this.getTop(), %this.getWidth(), %biggestY);
}

function Scratch::renderContainer(%this) {
	if(!isObject(%this)) return;
	if(!%this.container) return;
	// Variables
	%ws      = 4; // Whitespace between blocks.
	%x       = (%this.top ? 26 : (%this.blockType $= "INPUTBLOCK" ? 3 : 9));
	%biggest = 0;

	// First, go to our top connection if we have one.
	if(isObject(%obj = %this.topConnection)) {
		//%this.resize(%obj.getLeft()+mAbs(%obj.bottomConnector.getLeft()-%this.topConnector.getLeft()), %obj.getBottom()-7, %this.getWidth(), %this.getHeight());
		Scratch_Content.bringToFront(%this);
		Scratch::renderConnections(%this);
	} else Scratch::renderConnections(%this);

	// Or if we have a bottom connection, make that object come to us.
	if(isObject(%obj = %this.bottomConnection)) {
		%bc   = %this.bottomConnector;
		%tc   = %obj.topConnector;
		%newX = %this.getLeft()+(%bc.getLeft()-%tc.getLeft());
		if(%obj.getPosition() !$= %newX SPC %this.getBottom()-7) {
			%obj.resize(%newX, %this.getBottom()-7, %obj.getWidth(), %obj.getHeight());
			Scratch::update(%obj);
		}
	}

	// Push the control object first so that we can push arg boxes infront
	%this.pushToBack(%this.control);

	// Render our contents
	%ourH     = %this.container0.getHeight()-7;
	%newH     = getWord(%this.container0.minExtent, 1);
	for(%i=0;%i<%this.totalArgCount;%i++) {
		%child = %this.allArgObjects[%i];
		if(%child.ignore) continue; // .ignore is used mainly for background elements
		%lastChild = %child;
		%rectX     = (%child.static ? %child.getLeft() : %x); // If this object is static, then we'll only vertically align it.
		%rectY     = (%ourH/2)-(%child.getHeight()/2);
		if(%child.getPosition() $= %rectX SPC %rectY) continue; // This block is already correctly positioned.
		%child.resize(%rectX, %rectY, %child.getWidth(), %child.getHeight());
		if(%child.getHeight()+(%ws*2) > %newH) %newH = %child.getHeight()+(%this.bottom ? (%ws*2) : 0);
		%this.pushToBack(%child);
		%x += %child.getWidth()+%ws;
	}

	// Re-adjust the block size variables if neccessary
	if(isObject(%lastChild) && %lastChild.getRight()+%ws != %this.container0.getWidth()) {
		%x               = %lastChild.getRight()+%ws;
		%forceUpdateSize = 1;
	}

	// Resize ourselves
	if((%x > %this.container0.getWidth() || %newH != %ourH) || %forceUpdateSize) {
		//%newH  += (%this.bottom ? 7 : 0);
		%this.resize(%this.getLeft(), %this.getTop(), %x, (%newH+(%ws*2))+%this.container1.getHeight()+5);
		%this.container0.resize(%this.container0.getLeft(), %this.container0.getTop(), %x, %newH+(%ws*2));
	}

	// Render the arguments by argument index
	%count = -1;
	for(%i=0;%i<%this.getCount();%i++) {
		%object = %this.getObject(%i);
		if(%object.argIndex $= "") continue;
		%object[%object.argIndex] = %object;
		%count++;
	}
	for(%i=0;%i<%count+1;%i++) {
		%this.pushToBack(%object[%i]);
	}

	// Render our connections
	Scratch::renderContainerConnections(%this);

	// If we have inside top connection
	%ins_ourH = %this.getHeight()-7;
	%ins_newH = %this.container0.getHeight()+%this.container1.getHeight()-16;
	if(isObject(%obj = %this.inside_topConnection)) {
		%bc   = %this.container0.bottomConnector;
		%tc   = %obj.topConnector;
		%newX = %this.getLeft()+((%bc.getLeft()+%bc.getGroup().getLeft())-%tc.getLeft());
		if(%obj.getPosition() !$= %newX SPC (%this.container0.getBottom()+%this.getTop())-7) {
			%obj.resize(%newX, (%this.container0.getBottom()+%this.getTop())-7, %obj.getWidth(), %obj.getHeight());
			Scratch::update(%obj);
		}
		%btm = %obj;
		while(isObject(%btm)) {
			%ins_newH += %btm.getHeight();
			%btm       = %btm.bottomConnection;
		}
	} else %ins_newH += 16;

	if(%ins_newH != %ins_ourH) {
		%this.resize(%this.getLeft(), %this.getTop(), %this.getWidth(), %ins_newH);
	}

	// Resize the control
	%this.control.resize(0, 0, %this.container0.getWidth(), %this.container0.getHeight()-7);

	// Bring ourselves to the front
	%this.getGroup().bringToFront(%this);

	// Set our offsets
	Scratch::setBlockOffsets(%this);
}

function Scratch::renderContainerConnections(%this, %resetHeight, %doNotResize) {
	if(!isObject(%this)) return;
	// Render the inner connections first
	%ws       = 4;
	%biggestY = (%resetHeight ? getWord(%this.container0.minExtent, 1) : %this.container0.getHeight());
	if(!%this.renderedBefore) {
		%this.renderedBefore = 1;
		for(%i=0;%i<%this.getCount();%i++) {
			%child = %this.getObject(%i);
			if(%child.ignore) continue;
			if(Scratch::getBlockType(%child) !$= "ARG_PLACEHOLDER") continue;
			Scratch::setBlockOffsets(%child);
		}
	}
	for(%i=0;%i<getWordCount(%this.innerConnectionTable);%i++) {
		%object = getWord(%this.innerConnectionTable, %i);
		%child  = %object.innerConnection_Parent;
		Scratch::setBlockOffsets(%child);
		%object.resize(%this.getLeft()+%child.getLeft(), %this.getTop()+%child.getTop(), %object.getWidth(), %object.getHeight());
		%child.resize(%child.getLeft(), %child.getTop(), %object.getWidth(), %object.getHeight());
		if(%child.getHeight()+(%ws*2) > %biggestY) %biggestY = %child.getHeight()+(%ws*2);
		%object.getGroup().pushToBack(%object);
		Scratch::renderBlock(%object);
	}
	if(!%doNotResize) %this.container0.resize(%this.container0.getLeft(), %this.container0.getTop(), %this.container0.getWidth(), %biggestY);
}