if($CACHE::Scratch::SpawnSystem::CustomAmount $= "") $CACHE::Scratch::SpawnSystem::CustomAmount = 100;

function Scratch::spawnMenu_registerBlockType(%category, %label, %summary, %blockType, %blockArgs, %bitmap, %categoryColor) {
	// Clean the variables
	%blockType = atoi(%blockType);
	%blockArgs = %blockArgs;
	%label     = stripMLControlChars(trim(%label));
	%summary   = stripMLControlChars(trim(%summary));
	%category  = stripMLControlChars(trim(%category));

	// Create an objName identifier
	%ident = "Scratch_SpawnMenu_Button" @ getSafeVariableName(strLwr(%category)) @ "_" @ (%blockType == 34 ? $CACHE::Scratch::SpawnSystem::CustomAmount++ : getSafeVariableName(strLwr(%blockType))) @ "_" @ getSafeVariableName(strLwr(%blockArgs));

	// Validate the bitmap
	if(isFile(%bitmap @ ".PNG")) %validBmp = 1;
	else if(isFile(%bitmap @ ".JPG")) %validBmp = 1;
	else if(isFile(%bitmap @ ".JPEG")) %validBmp = 1;
	else if(isFile(%bitmap @ ".BMP")) %validBmp = 1;
	else %validBmp = 0;

	//Show any errors
	if(%label $= "" || %summary $= "" || %blockType < 0 || %category $= "") {
		echo("\c2ERROR: Scratch::spawnMenu_registerBlockType() - Invalid block info.");
		echo("\c2USAGE: Scratch::spawnMenu_registerBlockType(%category, %label, %summary, %blockType[, %blockArgs[, %bitmap[, %categoryColor]]])");
		backTrace();
		return 0;
	}

	// If this control already exists, then modify the existing one
	if(isObject(%obj = %ident)) {
		%obj.getObject(1).setBitmap((%validBmp ? %bitmap : "Add-Ons/Client_Scratch/img/blocks/unknown"));
		return 1;
	}

	// First, create the category
	Scratch::spawnMenu_createCategory(%category, %categoryColor);

	// Second, create the control
	%newObj = new GuiBitmapBorderCtrl(%ident) {
		profile = "ScratchSpawnButtonProfile";
		horizSizing = "right";
		vertSizing = "bottom";
		position = "0 0";
		extent = "96 96";
		minExtent = "8 2";
		visible = "1";
		blockType = %blockType;
		blockArgs = %blockArgs;
		category = %category;
		type = "spawnButton";
		label = %label;

		new GuiMLTextCtrl() {
				profile = "GuiMLTextProfile";
				horizSizing = "width";
				vertSizing = "top";
				position = "0 77";
				extent = "97 14";
				minExtent = "8 2";
				visible = "1";
				lineSpacing = "2";
				allowColorChars = "0";
				maxChars = "-1";
				text = "<just:center><font:Franklin Gothic Demi:14>" @ %label;
				maxBitmapHeight = "-1";
				selectable = "1";
		};
		new GuiBitmapCtrl() {
				profile = "GuiDefaultProfile";
				horizSizing = "width";
				vertSizing = "height";
				position = "0 0";
				extent = "96 96";
				minExtent = "8 2";
				visible = "1";
				bitmap = (%validBmp ? %bitmap : "Add-Ons/Client_Scratch/img/blocks/unknown");
				wrap = "0";
				lockAspectRatio = "1";
				alignLeft = "0";
				overflowImage = "0";
				keepCached = "0";
		};
		new GuiSwatchCtrl() {
				profile = "GuiDefaultProfile";
				horizSizing = "width";
				vertSizing = "top";
				position = "0 97";
				extent = "96 96";
				minExtent = "8 2";
				visible = "1";
				color = "255 255 255 230";

				new GuiMLTextCtrl() {
						profile = "GuiMLTextProfile";
						horizSizing = "width";
						vertSizing = "height";
						position = "6 3";
						extent = "85 42";
						minExtent = "8 2";
						visible = "1";
						lineSpacing = "2";
						allowColorChars = "0";
						maxChars = "-1";
						text = "<font:Franklin Gothic Demi:14>" @ %summary;
						maxBitmapHeight = "-1";
						selectable = "1";
				};
		};
		new GuiMouseEventCtrl(SpawnMenuButtonHandler) {
				profile = "GuiDefaultProfile";
				horizSizing = "width";
				vertSizing = "height";
				position = "0 0";
				extent = "96 96";
				minExtent = "8 2";
				visible = "1";
		};
	};
	%newObj.descCardObj = %newObj.getObject(2);
	%newObj.mouseEventCtrl = %newObj.getObject(3);
	Scratch_ProjectManager_SpawnMenuContent.add(%newObj);

	// Sort all items
	Scratch::spawnMenu_sortItems();
}

function Scratch::spawnMenu_createCategory(%catName, %catColor) {
	// Clean the variables
	%catName = stripMLControlChars(trim(%catName));
	if(stripChars(%catColor, "0123456789ABCDEF") !$= "" || strLen(%catColor) != 6) %catColor = "";

	// Run some validity checks
	if(%catName $= "") return 0;

	// Set some variables
	%objName = "Scratch_SpawnMenu_Category" @ getSafeVariableName(%catName);

	// Run more validity checks
	if(isObject(%objName)) return 1;

	// Create the control
	%newObj = new GuiSwatchCtrl(%objName) {
		profile = "GuiDefaultProfile";
		horizSizing = "width";
		vertSizing = "bottom";
		position = "0 0";
		extent = "609 26";
		minExtent = "8 2";
		visible = "1";
		color = "0 0 0 0";
		type = "category";
		category = %catName;

		new GuiMLTextCtrl() {
				profile = "GuiMLTextProfile";
				horizSizing = "width";
				vertSizing = "bottom";
				position = "0 2";
				extent = "609 16";
				minExtent = "8 2";
				visible = "1";
				lineSpacing = "2";
				allowColorChars = "0";
				maxChars = "-1";
				text = "<just:center><font:Franklin Gothic Demi:16>" @ (%catColor $= "" ? "" : "<color:" @ %catColor @ ">") @ %catName;
				maxBitmapHeight = "-1";
				selectable = "1";
		};
		new GuiBitmapCtrl() {
				profile = "GuiDefaultProfile";
				horizSizing = "width";
				vertSizing = "top";
				position = "10 16";
				extent = "589 10";
				minExtent = "8 2";
				visible = "1";
				bitmap = "Add-Ons/Client_Scratch/img/seperator";
				wrap = "1";
				lockAspectRatio = "0";
				alignLeft = "0";
				overflowImage = "0";
				keepCached = "0";
		};
	};
	Scratch_ProjectManager_SpawnMenuContent.add(%newObj);
	return 1;
}

function Scratch::spawnMenu_getSearchBar() {
	%this = Scratch_ProjectManager_SpawnMenuContent;
	if(isObject(%obj = $CACHE::Scratch::SpawnMenu::SearchBarObject)) return %obj;
	$CACHE::Scratch::SpawnMenu::SearchBarObject = new GuiTextEditCtrl() {
		profile     = "ScratchSpawnMenuSearchProfile";
		position    = "0 0";
		extent      = "20 18";
		horizSizing = "width";
		vertSizing  = "bottom";
		command     = "Scratch::spawnMenu_search();";
		text        = "Search...";
	};
	%this.add($CACHE::Scratch::SpawnMenu::SearchBarObject);
	return $CACHE::Scratch::SpawnMenu::SearchBarObject;
}

function Scratch::spawnMenu_sortItems() {
	Scratch::projMan_sortTabs();
	%this  = Scratch_ProjectManager_SpawnMenuContent;
	%bSize = 96; // Block size
	%ws    = 10; // Whitespace between elements
	%x     = 0;
	%y     = 0;
	%grpW  = %this.getWidth(); // Group width
	%grpH  = %this.getHeight(); // Group height
	%bigH  = 0; // Biggest height + Y Position -- Will be used to resize the container

	// Position the search bar
	%sBar  = Scratch::spawnMenu_getSearchBar();
	%sBar.resize(%x+%ws, %y+%ws, %grpW-((%x+%ws)*2), 18);
	%y    += 18+(%ws*2);
	%bigH  = %sBar.getBottom();

	// Generate a list of what we will render in order
	for(%i=0;%i<%this.getCount();%i++) {
		%object = %this.getObject(%i);
		if(%object.getId() == %sBar.getId()) continue; // We don't want to render the search bar; it's rendered by itself
		if(!%object.isVisible()) continue; // We don't want to render objects we cant see
		if(%object.type !$= "category") continue; // We dont want to render anything but categories
		%render[-1+%renderCount++] = %object;
		for(%j=0;%j<%this.getCount();%j++) {
			%child = %this.getObject(%j);
			if(!%child.isVisible()) continue; // We don't want to render objects we cant see
			if(%child.type !$= "spawnButton") continue;
			if(%child.category !$= %object.category) continue;
			%render[-1+%renderCount++] = %child;
		}
	}

	// Next, render the above list
	for(%i=0;%i<%renderCount;%i++) {
		%object = %render[%i];
		%objW   = %object.getWidth();
		%objH   = %object.getHeight();
		switch$(%object.type) {
			case "spawnButton":
				if(%x+%bSize >= %this.getWidth()) {
					%x  = 27;
					%y += %objH+%ws;
				}
				%object.resize(%x, %y, %bSize, %bSize);
				if(%object.getBottom() > %bigH) %bigH = %object.getBottom();
				%x += %bSize+%ws;
				
			case "category":
				if(%i != 0) %y += %bSize+%ws;
				%object.resize(0, %y, %grpW, %objH);
				if(%object.getBottom() > %bigH) %bigH = %object.getBottom();
				%y += %objH+%ws;
				%x  = 27;
				
			default: continue;
		}
	}
	if(%this.getHeight() != %bigH+%ws) {
		%this.resize(0, %this.getTop(), %this.getWidth(), %bigH+%ws);
	}
	Scratch::projMan_processSizes();
}

function Scratch::spawnMenu_search() {
	%group = Scratch_ProjectManager_SpawnMenuContent;
	%this  = Scratch::spawnMenu_getSearchBar();
	%query = %this.getValue();

	// If the user's query is empty, then clear the search bar and reset all spawn buttons
	if(%query $= "" || %query $= "Search..") {
		%this.setText("Search...");
		ScratchSpawnMenuSearchProfile.fontColor = "100 100 100 255";
		for(%j=0;%j<%group.getCount();%j++) {
			%object = %group.getObject(%j);
			%object.setVisible(1);
		}

		// Re-render the spawn menu
		Scratch::spawnMenu_sortItems();
		return;
	}

	// If the user is about to type, then remove the 'Search' string from the user's query
	if(getSubStr(%query, 0, 9) $= "Search...") {
		%this.setValue((%query = getSubStr(%query, 9, strLen(%query))));
		ScratchSpawnMenuSearchProfile.fontColor = "0 0 0 255";
	}

	if(getSubStr(%query, 1, 9) $= "Search...") {
		%query = getSubStr(%query, 0, 1) @ getSubStr(%query, 10, strLen(%query));
		%this.setValue(%query);
		ScratchSpawnMenuSearchProfile.fontColor = "0 0 0 255";
	}

	// Search for blocks that match the user's query
	for(%j=0;%j<%group.getCount();%j++) {
		%object = %group.getObject(%j);
		if(%object.getId() == %this.getId()) continue;
		if(%object.type !$= "spawnButton") continue;
		if(strIPos(%object.label, %query) != 0) { if(%object.isVisible()) %object.setVisible(0); continue; }
		if(!%object.isVisible()) %object.setVisible(1);
		%got[getStringCRC(%object.category)] = 1;
	}

	for(%j=0;%j<%group.getCount();%j++) {
		%object = %group.getObject(%j);
		if(%object.getId() == %this.getId()) continue;
		if(%object.type $= "spawnButton") continue;
		if(!%got[getStringCRC(%object.category)]) { %object.setVisible(0); continue; }
		%object.setVisible(1);
	}

	// Re-render the spawn menu
	Scratch::spawnMenu_sortItems();
}

function Scratch::spawnMenu_animateLoop(%this, %show, %loop) {
	if(!isObject(%this)) return;
	if(!isObject(%ctrl = %this.descCardObj)) return;
	%speed = 10;
	switch(%show) {
		case 0:
			if(%ctrl.getTop() >= 97) {
				%ctrl.resize(0, 97, 96, 96);
				return;
			}
			%ctrl.resize(0, %ctrl.getTop()+%speed, 96, 96);
		case 1:
			if(%ctrl.getTop() <= 0) {
				%ctrl.resize(0, 0, 96, 96);
				return;
			}
			%ctrl.resize(0, %ctrl.getTop()-%speed, 96, 96);
		default: return;
	}
	cancel($CACHE::Scratch::SpawnMenu::AnimSch[%this]);
	$CACHE::Scratch::SpawnMenu::AnimSch[%this] = schedule(10, 0, eval, "Scratch::spawnMenu_animateLoop(\"" @ %this @ "\", \"" @ %show @ "\", 1);");
}
// Mouse event handlers

function SpawnMenuButtonHandler::onMouseEnter(%this) {
	%group = %this.getGroup();
	Scratch::spawnMenu_animateLoop(%group, "1");
}

function SpawnMenuButtonHandler::onMouseLeave(%this) {
	%group = %this.getGroup();
	Scratch::spawnMenu_animateLoop(%group, "0");
}

function SpawnMenuButtonHandler::onMouseDown(%this) {
	%group = %this.getGroup();
	// Spawn the block
	%newObj = Scratch::createPremadeBlock(%group.blockType, %group.blockArgs);
	if(!isObject(%newObj)) {
		messageBoxOk("Internal error - Scratch Editor", "Unable to spawn \"" @ %group.label @ "\" (Scratch::createPremadeBlock did not return an object value)");
		return;
	}
	
	Scratch_Content.add(%newObj);
	Scratch::renderBlock(%newObj);
	Scratch::renderBlock(%newObj);
	Scratch::toggleProjectManager();
	if(!$Pref::Scratch::DoneWithTutorial && $CACHE::Scratch::Tutorial::Stage == 9) { $CACHE::Scratch::Tutorial::Stage++; Scratch::stepTutorial(); }
}

// Register default blocks

function initScratchSpawnMenu() {
	Scratch::spawnMenu_registerBlockType("Orange - Control Blocks", "Define Function", "A named block of code that you can later call for an indefinite amount of times.", "0", "",  "Add-Ons/Client_Scratch/img/blocks/defineFunc", "FF8C00");
	Scratch::spawnMenu_registerBlockType("Orange - Control Blocks", "Call Function",   "Call a function that has been defined.",                                           "1", "0", "Add-Ons/Client_Scratch/img/blocks/callFunc",   "FF8C00");
	Scratch::spawnMenu_registerBlockType("Orange - Control Blocks", "Return Block",    "Stop the current function and return a value.",                                    "7", "",  "Add-Ons/Client_Scratch/img/blocks/return",     "FF8C00");
	
	Scratch::spawnMenu_registerBlockType("Dark Red - Concatenation", "Normal",        "Combine two or more strings together.",                          "5",  "", "Add-Ons/Client_Scratch/img/blocks/concat_normal", "8B0000");
	Scratch::spawnMenu_registerBlockType("Dark Red - Concatenation", "Space",         "Combine two or more strings together, seperated by a space.",    "9",  "", "Add-Ons/Client_Scratch/img/blocks/concat_space",  "8B0000");
	Scratch::spawnMenu_registerBlockType("Dark Red - Concatenation", "Tab",           "Combine two or more strings together, seperated by a TAB.",      "10", "", "Add-Ons/Client_Scratch/img/blocks/concat_tab",    "8B0000");
	Scratch::spawnMenu_registerBlockType("Dark Red - Concatenation", "New Line",      "Combine two or more strings together, seperated by a new line.", "11", "", "Add-Ons/Client_Scratch/img/blocks/concat_nl",     "8B0000");
	Scratch::spawnMenu_registerBlockType("Dark Red - Concatenation", "Or Operation",  "Returns 'true' if one of two expressions evaluate to 'true'.",   "30", "", "Add-Ons/Client_Scratch/img/blocks/or",            "8B0000");
	Scratch::spawnMenu_registerBlockType("Dark Red - Concatenation", "And Operation", "Returns 'true' if two expressions evaluate to 'true'.",          "31", "", "Add-Ons/Client_Scratch/img/blocks/and",           "8B0000");
	
	Scratch::spawnMenu_registerBlockType("Red - Variables", "Local Variable",  "Reference a local variable. This does not persist between functions.", "3", "", "Add-Ons/Client_Scratch/img/blocks/localVar",  "FF0000");
	Scratch::spawnMenu_registerBlockType("Red - Variables", "Global Variable", "Reference a global variable. This persists until the game exits.",     "4", "", "Add-Ons/Client_Scratch/img/blocks/globalVar", "FF0000");
	
	Scratch::spawnMenu_registerBlockType("Blue - Variable Setting", "Normal (=)",          "Set a variable to a string.",                 "2",  "2", "Add-Ons/Client_Scratch/img/blocks/set_normal",   "0000FF");
	Scratch::spawnMenu_registerBlockType("Blue - Variable Setting", "Increment (++)",      "Increment a variable by one.",                "2",  "0", "Add-Ons/Client_Scratch/img/blocks/set_inc",      "0000FF");
	Scratch::spawnMenu_registerBlockType("Blue - Variable Setting", "Decrement (--)",      "Decrement a variable by one.",                "2",  "1", "Add-Ons/Client_Scratch/img/blocks/set_dec",      "0000FF");
	Scratch::spawnMenu_registerBlockType("Blue - Variable Setting", "Increase (+=)",       "Increment a variable by an amount.",          "2",  "3", "Add-Ons/Client_Scratch/img/blocks/set_incamt",   "0000FF");
	Scratch::spawnMenu_registerBlockType("Blue - Variable Setting", "Decrease (-=)",       "Increment a variable by an amount.",          "2",  "4", "Add-Ons/Client_Scratch/img/blocks/set_decamt",   "0000FF");
	Scratch::spawnMenu_registerBlockType("Blue - Variable Setting", "Multiply (*=)",       "Multiply a variable by an amount.",           "2",  "5", "Add-Ons/Client_Scratch/img/blocks/set_multiply", "0000FF");
	Scratch::spawnMenu_registerBlockType("Blue - Variable Setting", "Divide (/=)",         "Divide a variable by an amount.",             "2",  "6", "Add-Ons/Client_Scratch/img/blocks/set_divide",   "0000FF");
	Scratch::spawnMenu_registerBlockType("Blue - Variable Setting", "Modulate (%=)",       "Modulate a variable by an amount.",           "2",  "7", "Add-Ons/Client_Scratch/img/blocks/set_mod",      "0000FF");
	Scratch::spawnMenu_registerBlockType("Blue - Variable Setting", "Object Variable (=)", "Set a variable in a new object to a string.", "36", "",  "Add-Ons/Client_Scratch/img/blocks/set_objvar",   "0000FF");
	
	Scratch::spawnMenu_registerBlockType("Brown - Logical & Loops", "IF Block",      "Execute a block of code if an expression is true.",                        "12", "", "Add-Ons/Client_Scratch/img/blocks/if",      "A52A2A");
	Scratch::spawnMenu_registerBlockType("Brown - Logical & Loops", "FOR Block",     "Repeat a block of code an amount of times until an expression is true.",   "23", "", "Add-Ons/Client_Scratch/img/blocks/for",     "A52A2A");
	Scratch::spawnMenu_registerBlockType("Brown - Logical & Loops", "WHILE Block",   "Execute a block of code until an expression is false.",                    "25", "", "Add-Ons/Client_Scratch/img/blocks/while",   "A52A2A");
	Scratch::spawnMenu_registerBlockType("Brown - Logical & Loops", "New Object",    "Create an object with a classname and object name.",                       "17", "", "Add-Ons/Client_Scratch/img/blocks/newobj",  "A52A2A");
	Scratch::spawnMenu_registerBlockType("Brown - Logical & Loops", "Ternary Block", "If an expression is true, return a value, else return a different value.", "35", "", "Add-Ons/Client_Scratch/img/blocks/ternary", "A52A2A");
	
	Scratch::spawnMenu_registerBlockType("Green - Logical Comparisons", "Equal",                 "If a number and another number are the same, returns true. Else, returns false.",            "16", "0", "Add-Ons/Client_Scratch/img/blocks/logiccmp_equal",              "008800");
	Scratch::spawnMenu_registerBlockType("Green - Logical Comparisons", "Not Equal",             "If a number and another number are not the same, returns true. Else, returns false.",        "16", "1", "Add-Ons/Client_Scratch/img/blocks/logiccmp_notequal",           "008800");
	Scratch::spawnMenu_registerBlockType("Green - Logical Comparisons", "String Equal",          "If a string and another string are the same, returns true. Else, returns false.",            "16", "2", "Add-Ons/Client_Scratch/img/blocks/logiccmp_stringequal",        "008800");
	Scratch::spawnMenu_registerBlockType("Green - Logical Comparisons", "String Not Equal",      "If a string and another string are not the same, returns true. Else, returns false.",        "16", "3", "Add-Ons/Client_Scratch/img/blocks/logiccmp_stringnotequal",     "008800");
	Scratch::spawnMenu_registerBlockType("Green - Logical Comparisons", "Greater Than or Equal", "If a number is greater than or equal to another number, returns true. Else, returns false.", "16", "4", "Add-Ons/Client_Scratch/img/blocks/logiccmp_greaterthanorequal", "008800");
	Scratch::spawnMenu_registerBlockType("Green - Logical Comparisons", "Less Than or Equal",    "If a number is less than or equal to another number, returns true. Else, returns false.",    "16", "5", "Add-Ons/Client_Scratch/img/blocks/logiccmp_lessthanorequal",    "008800");
	Scratch::spawnMenu_registerBlockType("Green - Logical Comparisons", "Greater Than",          "If a number is greater than to another number, returns true. Else, returns false.",          "16", "6", "Add-Ons/Client_Scratch/img/blocks/logiccmp_greaterthan",        "008800");
	Scratch::spawnMenu_registerBlockType("Green - Logical Comparisons", "Less Than",             "If a number is greater than to another number, returns true. Else, returns false.",          "16", "7", "Add-Ons/Client_Scratch/img/blocks/logiccmp_lessthan",           "008800");
	Scratch::spawnMenu_registerBlockType("Green - Logical Comparisons", "Not",                   "Returns true if a string is not zero and is not empty.",                                     "16", "8", "Add-Ons/Client_Scratch/img/blocks/logiccmp_not",                "008800");
	
	Scratch::spawnMenu_registerBlockType("Purple - Arithmetic", "Add",      "Returns the sum of two numbers.",             "6", "0", "Add-Ons/Client_Scratch/img/blocks/arithmetic_add",      "800080");
	Scratch::spawnMenu_registerBlockType("Purple - Arithmetic", "Minus",    "Returns the difference between two numbers.", "6", "1", "Add-Ons/Client_Scratch/img/blocks/arithmetic_minus",    "800080");
	Scratch::spawnMenu_registerBlockType("Purple - Arithmetic", "Multiply", "Returns the product of two numbers.",         "6", "2", "Add-Ons/Client_Scratch/img/blocks/arithmetic_multiply", "800080");
	Scratch::spawnMenu_registerBlockType("Purple - Arithmetic", "Divide",   "Returns the quotient of two numbers.",        "6", "3", "Add-Ons/Client_Scratch/img/blocks/arithmetic_divide",   "800080");
	Scratch::spawnMenu_registerBlockType("Purple - Arithmetic", "Modulus",  "Returns the remainder of two numbers.",       "6", "4", "Add-Ons/Client_Scratch/img/blocks/arithmetic_modulo",   "800080");
	
	Scratch::spawnMenu_registerBlockType("Black - Package", "Package", "Returns the remainder of two numbers.", "20", "", "Add-Ons/Client_Scratch/img/blocks/package");
}
initScratchSpawnMenu();