if($CACHE::Scratch::SpawnSystem::CustomAmount $= "") $CACHE::Scratch::SpawnSystem::CustomAmount = 100;

function Scratch::workSpaces_registerSave(%category, %label, %summary, %saveFile, %bitmap) {
	// Clean the variables
	%blockType = atoi(%blockType);
	%blockArgs = %blockArgs;
	%label     = stripMLControlChars(trim(%label));
	%summary   = trim(%summary);
	%category  = stripMLControlChars(trim(%category));

	// Create an objName identifier
	%ident = "Scratch_workSpaces_Button" @ getSafeVariableName(strLwr(%category)) @ "_" @ getSafeVariableName(strLwr(%saveFile));

	// Validate the bitmap
	if(isFile(%bitmap @ ".PNG")) %validBmp = 1;
	else if(isFile(%bitmap @ ".JPG")) %validBmp = 1;
	else if(isFile(%bitmap @ ".JPEG")) %validBmp = 1;
	else if(isFile(%bitmap @ ".BMP")) %validBmp = 1;
	else %validBmp = 0;

	// Show any errors
	if(%label $= "" || %summary $= "" || %blockType < 0 || %category $= "") {
		echo("\c2ERROR: Scratch::workSpaces_registerBlockType() - Invalid block info.");
		echo("\c2USAGE: Scratch::workSpaces_registerBlockType(%category, %label, %summary, %saveFile[, %bitmap])");
		backTrace();
		return 0;
	}
	%bypass   = (getSubStr(%saveFile, 0, 2) $= "__");
	%saveFile = (%bypass ? getSubStr(%saveFile, 2, strLen(%saveFile)) : %saveFile);
	if(!isFile(%saveFile) && !%bypass) {
		echo("\c2ERROR: Scratch::workSpaces_registerBlockType() - Save file \"" @ %saveFile @ "\" does not exist.");
		return 0;
	}

	// If this control already exists, then modify the existing one
	if(isObject(%obj = %ident)) {
		%obj.getObject(1).setBitmap((%validBmp ? %bitmap : "Add-Ons/Client_Scratch/img/blocks/unknown"));
		return 1;
	}

	// First, create the category
	Scratch::workSpaces_createCategory(%category);

	// Second, create the control
	%newObj = new GuiBitmapBorderCtrl(%ident) {
		profile = "ScratchSpawnButtonProfile";
		horizSizing = "right";
		vertSizing = "bottom";
		position = "0 0";
		extent = "96 96";
		minExtent = "8 2";
		visible = "1";
		category = %category;
		type = "spawnButton";
		label = %label;
		saveFile = %saveFile;

		new GuiMLTextCtrl() {
				profile = "GuiMLTextProfile";
				horizSizing = "width";
				vertSizing = "top";
				position = "0 77";
				extent = "97 14";
				minExtent = "8 2";
				visible = "1";
				lineSpacing = "2";
				allowColorChars = "0";
				maxChars = "-1";
				text = "<just:center><font:Franklin Gothic Demi:14>" @ %label;
				maxBitmapHeight = "-1";
				selectable = "1";
		};
		new GuiBitmapCtrl() {
				profile = "GuiDefaultProfile";
				horizSizing = "width";
				vertSizing = "height";
				position = "0 0";
				extent = "96 96";
				minExtent = "8 2";
				visible = "1";
				bitmap = (%validBmp ? %bitmap : "Add-Ons/Client_Scratch/img/blocks/unknown");
				wrap = "0";
				lockAspectRatio = "1";
				alignLeft = "0";
				overflowImage = "0";
				keepCached = "0";
		};
		new GuiSwatchCtrl() {
				profile = "GuiDefaultProfile";
				horizSizing = "width";
				vertSizing = "top";
				position = "0 97";
				extent = "96 96";
				minExtent = "8 2";
				visible = "1";
				color = "255 255 255 230";

				new GuiMLTextCtrl() {
						profile = "GuiMLTextProfile";
						horizSizing = "width";
						vertSizing = "height";
						position = "6 3";
						extent = "85 42";
						minExtent = "8 2";
						visible = "1";
						lineSpacing = "2";
						allowColorChars = "0";
						maxChars = "-1";
						text = "<font:Franklin Gothic Demi:14>" @ %summary;
						maxBitmapHeight = "-1";
						selectable = "1";
				};
		};
		new GuiMouseEventCtrl(WorkspaceSaveButtonHandler) {
				profile = "GuiDefaultProfile";
				horizSizing = "width";
				vertSizing = "height";
				position = "0 0";
				extent = "96 96";
				minExtent = "8 2";
				visible = "1";
		};
	};
	%newObj.descCardObj = %newObj.getObject(2);
	%newObj.mouseEventCtrl = %newObj.getObject(3);
	Scratch_ProjectManager_Workspaces.add(%newObj);

	// Sort all items
	Scratch::workSpaces_sortItems();
}

function Scratch::workSpaces_createCategory(%catName) {
	// Clean the variables
	%catName = stripMLControlChars(trim(%catName));

	// Run some validity checks
	if(%catName $= "") return 0;

	// Set some variables
	%objName = "Scratch_workSpaces_Category" @ getSafeVariableName(%catName);

	// Run more validity checks
	if(isObject(%objName)) return 1;

	// Create the control
	%newObj = new GuiSwatchCtrl(%objName) {
		profile = "GuiDefaultProfile";
		horizSizing = "width";
		vertSizing = "bottom";
		position = "0 0";
		extent = "609 26";
		minExtent = "8 2";
		visible = "1";
		color = "0 0 0 0";
		type = "category";
		category = %catName;

		new GuiMLTextCtrl() {
				profile = "GuiMLTextProfile";
				horizSizing = "width";
				vertSizing = "bottom";
				position = "0 2";
				extent = "609 16";
				minExtent = "8 2";
				visible = "1";
				lineSpacing = "2";
				allowColorChars = "0";
				maxChars = "-1";
				text = "<just:center><font:Franklin Gothic Demi:16>" @ %catName;
				maxBitmapHeight = "-1";
				selectable = "1";
		};
		new GuiBitmapCtrl() {
				profile = "GuiDefaultProfile";
				horizSizing = "width";
				vertSizing = "top";
				position = "10 16";
				extent = "589 10";
				minExtent = "8 2";
				visible = "1";
				bitmap = "Add-Ons/Client_Scratch/img/seperator";
				wrap = "1";
				lockAspectRatio = "0";
				alignLeft = "0";
				overflowImage = "0";
				keepCached = "0";
		};
	};
	Scratch_ProjectManager_Workspaces.add(%newObj);
	return 1;
}

function Scratch::workSpaces_sortItems() {
	Scratch::projMan_sortTabs();
	%this  = Scratch_ProjectManager_Workspaces;
	%bSize = 96; // Block size
	%ws    = 10; // Whitespace between elements
	%x     = 0;
	%y     = 0;
	%grpW  = %this.getWidth(); // Group width
	%grpH  = %this.getHeight(); // Group height
	%bigH  = 0; // Biggest height + Y Position -- Will be used to resize the container

	// Generate a list of what we will render in order
	for(%i=0;%i<%this.getCount();%i++) {
		%object = %this.getObject(%i);
		if(%object.type !$= "category") continue;
		%render[-1+%renderCount++] = %object;
		for(%j=0;%j<%this.getCount();%j++) {
			%child = %this.getObject(%j);
			if(%child.type !$= "spawnButton") continue;
			if(%child.category !$= %object.category) continue;
			%render[-1+%renderCount++] = %child;
		}
	}

	// Next, render the above list
	for(%i=0;%i<%renderCount;%i++) {
		%object = %render[%i];
		%objW   = %object.getWidth();
		%objH   = %object.getHeight();
		switch$(%object.type) {
			case "spawnButton":
				if(%x+%bSize >= %this.getWidth()) {
					%x  = 27;
					%y += %objH+%ws;
				}
				%object.resize(%x, %y, %bSize, %bSize);
				if(%object.getBottom() > %bigH) %bigH = %object.getBottom();
				%x += %bSize+%ws;
				
			case "category":
				if(%i != 0) %y += %bSize+%ws;
				%object.resize(0, %y, %grpW, %objH);
				if(%object.getBottom() > %bigH) %bigH = %object.getBottom();
				%y += %objH+%ws;
				%x  = 27;
				
			default: continue;
		}
	}
	if(%this.getHeight() != %bigH+%ws) {
		%this.resize(0, %this.getTop(), %this.getWidth(), %bigH+%ws);
	}
	Scratch::projMan_processSizes();
}

function Scratch::workSpaces_animateLoop(%this, %show, %loop) {
	if(!isObject(%this)) return;
	if(!isObject(%ctrl = %this.descCardObj)) return;
	%speed = 10;
	switch(%show) {
		case 0:
			if(%ctrl.getTop() >= 97) {
				%ctrl.resize(0, 97, 96, 96);
				return;
			}
			%ctrl.resize(0, %ctrl.getTop()+%speed, 96, 96);
		case 1:
			if(%ctrl.getTop() <= 0) {
				%ctrl.resize(0, 0, 96, 96);
				return;
			}
			%ctrl.resize(0, %ctrl.getTop()-%speed, 96, 96);
		default: return;
	}
	cancel($CACHE::Scratch::SpawnMenu::AnimSch[%this]);
	$CACHE::Scratch::SpawnMenu::AnimSch[%this] = schedule(10, 0, eval, "Scratch::workSpaces_animateLoop(\"" @ %this @ "\", \"" @ %show @ "\", 1);");
}

function Scratch::workSpaces_deleteSave(%file, %confirm) {
	if(!isFile(%file)) {
		messageBoxOk("Error - Scratch Editor", "Scratch::workSpaces_deleteSave() failed:\n\n\"File '" @ %file @ "' does not exist\"");
		error("ERROR: Scratch::workSpaces_deleteSave() - File '" @ %file @ "' does not exist.");
		return;
	}
	if(fileExt(%file) !$= ".sws") {
		error("ERROR: Scratch::workSpaces_deleteSave() - File '" @ %file @ "' does not have the '.sws' filetype.");
		return;
	}
	if(!%confirm) {
		messageBoxYesNo("Confirmation - Scratch Editor", "Are you sure you want to delete '" @ %file @ "'?", "Scratch::workSpaces_deleteSave(\"" @ %file @ "\", 1);");
		return;
	}
	fileDelete(%file);
	initScratchWorkspaceMenu();
}
// Mouse event handlers

function WorkspaceSaveButtonHandler::onRightMouseDown(%this) {
	%group  = %this.getGroup();
	%file   = %group.saveFile;
	if(%file $= "SAVE") return;
	if(%file $= "SAVEAS") return;
	if(%file $= "REFRESH") return;
	%string =           "<font:Arial Bold:21>" @ fileBase(%file);
	%string = %string @ "<br><font:Arial Italic:12>" @ %file @ "<font:Arial:14>";
	%string = %string @ "<br>Last modified " @ Scratch::getDateDifference(getFileModifiedTime(%file)) @ " ago";
	messageBoxYesNo("Info - Scratch Editor", %string @ "\n\nDelete this save?", "Scratch::workSpaces_deleteSave(\"" @ %this.getGroup().saveFile @ "\");");
}

function WorkspaceSaveButtonHandler::onMouseEnter(%this) {
	%group = %this.getGroup();
	Scratch::workSpaces_animateLoop(%group, "1");
}

function WorkspaceSaveButtonHandler::onMouseLeave(%this) {
	%group = %this.getGroup();
	Scratch::workSpaces_animateLoop(%group, "0");
}

function Scratch_WorkspaceSaveAsCallback(%path, %file) {
	Scratch::saveWorkspace(%path @ "/" @ %file);
	setModPaths(getModPaths());
	initScratchWorkspaceMenu();
}

function WorkspaceSaveButtonHandler::onMouseDown(%this) {
	%group = %this.getGroup();
	%file  = %group.saveFile;
	switch$(%file) {
		case "SAVE":
			Scratch::saveWorkspace("config/client/Scratch/saves/" @ strReplace(getWord(getDateTime(), 0), "/", "-") @ "/" @ strReplace(getWord(getDateTime(), 1), ":", ";") @ ".sws", strReplace(getWord(getDateTime(), 1), ":", ";"));
			setModPaths(getModPaths());
			initScratchWorkspaceMenu();
		case "SAVEAS": ScratchGui.openDirectorySelector("Scratch_WorkspaceSaveAsCallback", "Save", "_.sws_ Scratch Workspace (.sws)");
		case "REFRESH": initScratchWorkspaceMenu(); return;
		default:
			messageBoxYesNo("Confirmation - Scratch", "Are you sure you want to load \"" @ %file @ "\"?\n\nYou will lose all unsaved progress on the current workspace.", "Scratch::toggleProjectManager();Scratch::loadWorkspace(" @ getSafeArgument(%file) @ ");");
			return;
	}
	Scratch::toggleProjectManager();
}

// Init the menu

function Scratch::subtractDates(%strOne, %strTwo) {
	%dateOne  = strReplace(getWord(%strOne, 0), "/", " ");
	%dateTwo  = strReplace(getWord(%strTwo, 0), "/", " ");
	%timeOne  = strReplace(getWord(%strOne, 2), ":", " ");
	%timeTwo  = strReplace(getWord(%strTwo, 2), ":", " ");

	%monDif   = mAbs(getWord(%dateTwo, 0) - getWord(%dateOne, 0));
	%dayDif   = mAbs(getWord(%dateTwo, 1) - getWord(%dateOne, 1));
	%yearDif  = mAbs(getWord(%dateTwo, 2) - getWord(%dateOne, 2));

	%hourDif  = mAbs(getWord(%timeTwo, 0) - getWord(%timeOne, 0));
	%minsDif  = mAbs(getWord(%timeTwo, 1) - getWord(%timeOne, 1));
	%secsDif  = mAbs(getWord(%timeTwo, 2) - getWord(%timeOne, 2));

	if(%yearDif != 0) %str = (%str $= "" ? "" : %str @ ", " ) @ %yearDif @ " year" @ (%yearDif == 1 ? "" : "s");
	if(%monDif != 0)  %str = (%str $= "" ? "" : %str @ ", " ) @ %monDif @ " month" @ (%monDif == 1 ? "" : "s");
	if(%dayDif != 0)  %str = (%str $= "" ? "" : %str @ ", " ) @ %dayDif @ " day" @ (%dayDif == 1 ? "" : "s");

	if(%hourDif != 0) %str = (%str $= "" ? "" : %str @ ", " ) @ %hourDif @ " hour" @ (%hourDif == 1 ? "" : "s");
	if(%minsDif != 0) %str = (%str $= "" ? "" : %str @ ", " ) @ %minsDif @ " minute" @ (%minsDif == 1 ? "" : "s");
	if(%secsDif != 0) %str = (%str $= "" ? "" : %str @ ", " ) @ %secsDif @ " second" @ (%secsDif == 1 ? "" : "s");

	return %str;
}

function Scratch::getDateDifference(%date) {
	%currDate = strReplace(trim(strReplace(getDateTime(), "0", "\t")), "\t", "0");
	%currDate = getWord(%currDate, 0) SPC "-" SPC getWord(%currDate, 1);
	%str      = Scratch::subtractDates(%currDate, %date);
	return %str;
}

function initScratchWorkspaceMenu() {
	Scratch_ProjectManager_Workspaces.deleteAll();
	%ptrn = "*.sws"; // .sws = Scratch WorkSpace

	%SO = new FileObject();

	for(%file=findFirstFile(%ptrn);%file !$= "";%file=findNextFile(%ptrn)) {
		%summary = "<font:Franklin Gothic Demi:19>" @ fileName(%file) @ "<font:Franklin Gothic Demi:14>";
		
		// Read the file to get the block count
		%blockCount = Scratch::getWorkspaceBlockCount(%file);
		
		%summary = %summary @ "<br>" @ Scratch::neatifyNumber(%blockCount) @ " block" @ (%blockCount == 1 ? "" : "s");
		%summary = %summary @ "<br>" @ getFileModifiedTime(%file);

		Scratch::workSpaces_registerSave("Saves", fileBase(%file), %summary, %file, filePath(%file) @ "/" @ fileBase(%file));
	}
	Scratch::workSpaces_registerSave("File Options", "Save",       "Save the current workspace.",           "__SAVE",    "Add-Ons/Client_Scratch/img/saveWorkspace");
	Scratch::workSpaces_registerSave("File Options", "Save As...", "Save the current workspace to a file.", "__SAVEAS",  "Add-Ons/Client_Scratch/img/saveWorkspace");
	Scratch::workSpaces_registerSave("File Options", "Refresh",    "Refresh the listing.",                  "__REFRESH", "Add-Ons/Client_Scratch/img/refresh");
	%SO.delete();
}